"""
Partially linear controllers.

Partially linear controllers are encoded as sets of linear controllers and
anchor points. The anchors are coordinates in the state space. For each state,
the linear controller with the closest anchor point is used. In other words,
these controllers are basically choices among multiple
:mod:`~moptipyapps.dynamic_control.controllers.linear` controllers.
"""

from typing import Final, Iterable

import numba  # type: ignore
import numpy as np

from moptipyapps.dynamic_control.controller import Controller
from moptipyapps.dynamic_control.system import System


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __linear_2d_1o_2(state: np.ndarray, _: float,
                     params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a double linear function for 2d state spaces.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]

    d: float = ((s0 - params[0]) ** 2.0) + ((s1 - params[1]) ** 2.0)
    o: float = (s0 * params[2]) + (s1 * params[3])

    d2: float = ((s0 - params[4]) ** 2.0) + ((s1 - params[5]) ** 2.0)
    if d2 < d:
        o = (s0 * params[6]) + (s1 * params[7])
    out[0] = o


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __linear_3d_1o_2(state: np.ndarray, _: float,
                     params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a double linear function for 3d state spaces.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    s2: Final[float] = state[2]

    d: float = (((s0 - params[0]) ** 2.0) + ((s1 - params[1]) ** 2.0)
                + ((s2 - params[2]) ** 2.0))
    o: float = (s0 * params[3]) + (s1 * params[4]) + (s2 * params[5])

    d2: float = (((s0 - params[6]) ** 2.0) + ((s1 - params[7]) ** 2.0)
                 + ((s2 - params[8]) ** 2.0))
    if d2 < d:
        o = (s0 * params[9]) + (s1 * params[10]) + (s2 * params[11])

    out[0] = o


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __linear_2d_1o_3(state: np.ndarray, _: float,
                     params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a triple linear function for 2d state spaces.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]

    d: float = ((s0 - params[0]) ** 2.0) + ((s1 - params[1]) ** 2.0)
    o: float = (s0 * params[2]) + (s1 * params[3])

    d2: float = ((s0 - params[4]) ** 2.0) + ((s1 - params[5]) ** 2.0)
    if d2 < d:
        o = (s0 * params[6]) + (s1 * params[7])

    d2 = ((s0 - params[8]) ** 2.0) + ((s1 - params[9]) ** 2.0)
    if d2 < d:
        o = (s0 * params[10]) + (s1 * params[11])

    out[0] = o


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __linear_3d_1o_3(state: np.ndarray, _: float,
                     params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a double linear function for 3d state spaces.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    s2: Final[float] = state[2]

    d: float = (((s0 - params[0]) ** 2.0) + ((s1 - params[1]) ** 2.0)
                + ((s2 - params[2]) ** 2.0))
    o: float = (s0 * params[3]) + (s1 * params[4]) + (s2 * params[5])

    d2: float = (((s0 - params[6]) ** 2.0) + ((s1 - params[7]) ** 2.0)
                 + ((s2 - params[8]) ** 2.0))
    if d2 < d:
        o = (s0 * params[9]) + (s1 * params[10]) + (s2 * params[11])

    d2 = (((s0 - params[12]) ** 2.0) + ((s1 - params[13]) ** 2.0)
          + ((s2 - params[14]) ** 2.0))
    if d2 < d:
        o = (s0 * params[15]) + (s1 * params[16]) + (s2 * params[17])

    out[0] = o


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __linear_2d_1o_4(state: np.ndarray, _: float,
                     params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a quadruple linear function for 2d state spaces.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]

    d: float = ((s0 - params[0]) ** 2.0) + ((s1 - params[1]) ** 2.0)
    o: float = (s0 * params[2]) + (s1 * params[3])

    d2: float = ((s0 - params[4]) ** 2.0) + ((s1 - params[5]) ** 2.0)
    if d2 < d:
        o = (s0 * params[6]) + (s1 * params[7])

    d2 = ((s0 - params[8]) ** 2.0) + ((s1 - params[9]) ** 2.0)
    if d2 < d:
        o = (s0 * params[10]) + (s1 * params[11])

    d2 = ((s0 - params[12]) ** 2.0) + ((s1 - params[13]) ** 2.0)
    if d2 < d:
        o = (s0 * params[14]) + (s1 * params[15])

    out[0] = o


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __linear_3d_1o_4(state: np.ndarray, _: float,
                     params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a quadruple linear function for 3d state spaces.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    s2: Final[float] = state[2]

    d: float = (((s0 - params[0]) ** 2.0) + ((s1 - params[1]) ** 2.0)
                + ((s2 - params[2]) ** 2.0))
    o: float = (s0 * params[3]) + (s1 * params[4]) + (s2 * params[5])

    d2: float = (((s0 - params[6]) ** 2.0) + ((s1 - params[7]) ** 2.0)
                 + ((s2 - params[8]) ** 2.0))
    if d2 < d:
        o = (s0 * params[9]) + (s1 * params[10]) + (s2 * params[11])

    d2 = (((s0 - params[12]) ** 2.0) + ((s1 - params[13]) ** 2.0)
          + ((s2 - params[14]) ** 2.0))
    if d2 < d:
        o = (s0 * params[15]) + (s1 * params[16]) + (s2 * params[17])

    d2 = (((s0 - params[18]) ** 2.0) + ((s1 - params[19]) ** 2.0)
          + ((s2 - params[20]) ** 2.0))
    if d2 < d:
        o = (s0 * params[21]) + (s1 * params[22]) + (s2 * params[23])

    out[0] = o


def partially_linear(system: System) -> Iterable[Controller]:
    """
    Create a several linear controllers for the given equations object.

    :param system: the equations object
    :return: the partially linear controllers
    """
    if system.control_dims != 1:
        raise ValueError("invalid controller dimensions "
                         f"{system.control_dims} for {system!r}.")
    if system.state_dims == 2:
        return (Controller("linear_2", 2, 1, 8, __linear_2d_1o_2),
                Controller("linear_3", 2, 1, 12, __linear_2d_1o_3),
                Controller("linear_4", 2, 1, 16, __linear_2d_1o_4))
    if system.state_dims == 3:
        return (Controller("linear_2", 3, 1, 12, __linear_3d_1o_2),
                Controller("linear_3", 3, 1, 18, __linear_3d_1o_3),
                Controller("linear_4", 3, 1, 24, __linear_3d_1o_4))
    raise ValueError("invalid state dimensions "
                     f"{system.state_dims} for {system!r}.")



"""Some shared variables and constants."""

import argparse
from typing import Any, Callable, Final

import moptipy.examples.jssp.instance as ins
from pycommons.io.arguments import make_argparser, make_epilog

from moptipyapps.version import __version__ as moptipyapps_version

#: the instance scope
SCOPE_INSTANCE: Final[str] = ins.SCOPE_INSTANCE


def moptipyapps_argparser(file: str, description: str,
                          epilog: str) -> argparse.ArgumentParser:
    """
    Create an argument parser with default settings.

    :param file: the `__file__` special variable of the calling script
    :param description: the description string
    :param epilog: the epilogue string
    :returns: the argument parser

    >>> ap = moptipyapps_argparser(
    ...     __file__, "This is a test program.", "This is a test.")
    >>> isinstance(ap, argparse.ArgumentParser)
    True
    >>> "Copyright" in ap.epilog
    True
    """
    return make_argparser(
        file, description,
        make_epilog(epilog, 2023, 2024, "Thomas Weise",
                    url="https://thomasweise.github.io/moptipyapps",
                    email="tweise@hfuu.edu.cn, tweise@ustc.edu.cn"),
        moptipyapps_version)


def motipyapps_footer_bottom_comments(
        _: Any, dest: Callable[[str], Any],
        additional: str | None = None) -> None:
    """
    Print the standard csv footer for moptipyapps.

    :param _: the setup object, ignored
    :param dest: the destination callable
    :param dest: the destination to write to
    :param additional: any additional output string

    >>> def __qpt(s: str):
    ...     print(s[:49])
    >>> motipyapps_footer_bottom_comments(None, __qpt, "bla")
    This data has been generated with moptipyapps ver
    bla
    You can find moptipyapps at https://thomasweise.g

    >>> motipyapps_footer_bottom_comments(None, __qpt, None)
    This data has been generated with moptipyapps ver
    You can find moptipyapps at https://thomasweise.g
    """
    dest("This data has been generated with moptipyapps version "
         f"{moptipyapps_version}.")
    if (additional is not None) and (str.__len__(additional) > 0):
        dest(additional)
    dest("You can find moptipyapps at "
         "https://thomasweise.github.io/moptipyapps.")



"""Perform tests on the Two-Dimensional Bin Packing Problem."""

from time import monotonic_ns
from typing import Callable, Final, Iterable, cast

import numpy as np
import numpy.random as rnd
from moptipy.api.algorithm import Algorithm
from moptipy.api.encoding import Encoding
from moptipy.api.objective import Objective
from moptipy.operators.signed_permutations.op0_shuffle_and_flip import (
    Op0ShuffleAndFlip,
)
from moptipy.spaces.signed_permutations import SignedPermutations
from moptipy.tests.algorithm import validate_algorithm
from moptipy.tests.encoding import validate_encoding
from moptipy.tests.objective import validate_objective
from moptipy.tests.space import validate_space
from numpy.random import Generator, default_rng
from pycommons.types import type_error

from moptipyapps.binpacking2d.encodings.ibl_encoding_1 import (
    ImprovedBottomLeftEncoding1,
)
from moptipyapps.binpacking2d.encodings.ibl_encoding_2 import (
    ImprovedBottomLeftEncoding2,
)
from moptipyapps.binpacking2d.instance import Instance
from moptipyapps.binpacking2d.objectives.bin_count_and_last_empty import (
    BinCountAndLastEmpty,
)
from moptipyapps.binpacking2d.packing import Packing
from moptipyapps.binpacking2d.packing_space import PackingSpace

#: the internal random number generator
__RANDOM: Final[Generator] = default_rng()


def binpacking_instances_for_tests(
        random: Generator = __RANDOM) -> Iterable[str]:
    """
    Get a sequence of 2D Bin Packing instances to test on.

    :param random: the random number generator to use
    :returns: an iterable of 2D Bin Packing instance names
    """
    ri = random.integers
    insts: set[str] = {
        "a04", "a08", "beng10", f"a0{ri(1, 10)}", f"a1{ri(1, 10)}",
        f"a2{ri(1, 10)}", f"a3{ri(1, 10)}", f"a4{ri(1, 4)}",
        f"beng0{ri(1, 10)}", f"cl01_080_0{ri(1, 10)}",
        f"cl06_020_0{ri(1, 10)}", f"cl09_040_0{ri(1, 10)}",
        f"cl10_100_0{ri(1, 10)}"}
    insn: list[str] = list(Instance.list_resources())
    while len(insts) < 16:
        insts.add(insn.pop(ri(len(insn))))
    use_insts: list[str] = list(insts)
    random.shuffle(cast(list, use_insts))
    return use_insts


def make_packing_valid(inst: Instance,
                       random: Generator = __RANDOM) \
        -> Callable[[Generator, Packing], Packing]:
    """
    Make a function that creates valid packings.

    :param inst: the two-dimensional bin packing instance
    :param random: the random number generator to use
    :returns: a function that can make packings valid
    """
    search_space = SignedPermutations(inst.get_standard_item_sequence())
    encoding = (ImprovedBottomLeftEncoding1 if random.integers(2) == 0
                else ImprovedBottomLeftEncoding2)(inst)
    op0 = Op0ShuffleAndFlip(search_space)

    def __make_valid(ra: rnd.Generator,
                     y: Packing, ss=search_space,
                     en=encoding, o0=op0) -> Packing:
        x = ss.create()
        o0.op0(ra, x)
        en.decode(x, y)
        return y

    return __make_valid


def make_packing_invalid(random: Generator = __RANDOM) \
        -> Callable[[Packing], Packing]:
    """
    Make a function that creates invalid packings.

    :param random: the random number generator to use
    :returns: a function that can make packings invalid
    """

    def __make_invalid(x: Packing, ri=random.integers) -> Packing:
        not_finished: bool = True
        end_time: Final[int] = monotonic_ns() + 20_000_000_000
        while not_finished:
            while ri(2) == 0:
                if monotonic_ns() >= end_time:
                    x[0, 0] = -1
                    return x
                x[ri(len(x)), ri(6)] = -1
                not_finished = False
            while ri(2) == 0:
                second = first = ri(len(x))
                while second == first:
                    if monotonic_ns() >= end_time:
                        x[0, 0] = -1
                        return x
                    second = ri(len(x))
                x[first, 1] = x[second, 1]
                x[first, 2] = x[second, 2] - 1
                x[first, 3] = x[second, 3] - 1
                x[first, 4] = x[second, 4] + 1
                x[first, 5] = x[second, 5] + 1
        return x

    return __make_invalid


def validate_algorithm_on_1_2dbinpacking(
        algorithm: Algorithm | Callable[
            [Instance, SignedPermutations, Objective], Algorithm],
        instance: str | None = None, max_fes: int = 100,
        random: Generator = __RANDOM) -> None:
    """
    Check the validity of a black-box algorithm on the 2d bin packing.

    :param algorithm: the algorithm or algorithm factory
    :param instance: the instance name, or `None` to randomly pick one
    :param max_fes: the maximum number of FEs
    :param random: the default random generator to use
    """
    if not (isinstance(algorithm, Algorithm) or callable(algorithm)):
        raise type_error(algorithm, "algorithm", Algorithm, True)
    if instance is None:
        instance = str(random.choice(Instance.list_resources()))
    if not isinstance(instance, str):
        raise type_error(instance, "bin packing instance name", (str, None))
    inst = Instance.from_resource(instance)
    if not isinstance(inst, Instance):
        raise type_error(inst, f"loaded bin packing instance {instance!r}",
                         Instance)

    search_space = SignedPermutations(inst.get_standard_item_sequence())
    solution_space = PackingSpace(inst)
    encoding = (ImprovedBottomLeftEncoding1 if random.integers(2) == 0
                else ImprovedBottomLeftEncoding2)(inst)
    objective = BinCountAndLastEmpty(inst)
    if callable(algorithm):
        algorithm = algorithm(inst, search_space, objective)
    if not isinstance(algorithm, Algorithm):
        raise type_error(algorithm, "algorithm", Algorithm, call=True)

    validate_algorithm(algorithm=algorithm,
                       solution_space=solution_space,
                       objective=objective,
                       search_space=search_space,
                       encoding=encoding,
                       max_fes=max_fes)


def validate_algorithm_on_2dbinpacking(
        algorithm: Callable[[Instance, SignedPermutations,
                             Objective], Algorithm],
        max_fes: int = 100, random: Generator = __RANDOM) -> None:
    """
    Validate an algorithm on a set of bin packing instances.

    :param algorithm: the algorithm factory
    :param max_fes: the maximum FEs
    :param random: the random number generator
    """
    end_time: Final[int] = monotonic_ns() + 20_000_000_000
    for i in binpacking_instances_for_tests(random):
        if monotonic_ns() >= end_time:
            break
        validate_algorithm_on_1_2dbinpacking(algorithm, i, max_fes, random)


def validate_objective_on_1_2dbinpacking(
        objective: Objective | Callable[[Instance], Objective],
        instance: str | None = None,
        random: Generator = __RANDOM) -> None:
    """
    Validate an objective function on 1 2D bin packing instance.

    :param objective: the objective function or a factory creating it
    :param instance: the instance name
    :param random: the random number generator
    """
    if instance is None:
        instance = str(random.choice(Instance.list_resources()))
    if not isinstance(instance, str):
        raise type_error(instance, "bin packing instance name", (str, None))
    inst = Instance.from_resource(instance)
    if not isinstance(inst, Instance):
        raise type_error(inst, f"loaded bin packing instance {instance!r}",
                         Instance)

    if callable(objective):
        objective = objective(inst)

    validate_objective(
        objective=objective,
        solution_space=PackingSpace(inst),
        make_solution_space_element_valid=make_packing_valid(inst),
        is_deterministic=True)


def validate_objective_on_2dbinpacking(
        objective: Objective | Callable[[Instance], Objective],
        random: Generator = __RANDOM) -> None:
    """
    Validate an objective function on bin packing instances.

    :param objective: the objective function or a factory creating it
    :param random: the random number generator
    """
    end_time: Final[int] = monotonic_ns() + 20_000_000_000
    for i in binpacking_instances_for_tests(random):
        if monotonic_ns() >= end_time:
            break
        validate_objective_on_1_2dbinpacking(objective, i, random)


def validate_signed_permutation_encoding_on_1_2dbinpacking(
        encoding: Encoding | Callable[[Instance], Encoding],
        instance: str | None = None,
        random: Generator = __RANDOM) -> None:
    """
    Validate a signed permutation encoding on one 2D bin packing instance.

    :param encoding: the encoding or a factory creating it
    :param instance: the instance name
    :param random: the random number generator
    """
    if instance is None:
        instance = str(random.choice(Instance.list_resources()))
    if not isinstance(instance, str):
        raise type_error(instance, "bin packing instance name", (str, None))
    inst = Instance.from_resource(instance)
    if not isinstance(inst, Instance):
        raise type_error(inst, f"loaded bin packing instance {instance!r}",
                         Instance)

    if callable(encoding):
        encoding = encoding(inst)
    inst = Instance.from_resource(instance)

    x_space = SignedPermutations(inst.get_standard_item_sequence())
    validate_space(x_space)

    y_space = PackingSpace(inst)
    validate_space(y_space, make_element_valid=None)

    validate_encoding(encoding, x_space, y_space)

    x = x_space.create()
    x_space.validate(x)

    y = y_space.create()
    encoding.decode(x, y)
    y_space.validate(y)

    random.shuffle(x)
    ri = random.integers
    for i, xx in enumerate(x):
        if ri(2) == 0:
            x[i] = -xx
    encoding.decode(x, y)
    y_space.validate(y)

    x_str = x_space.to_str(x)
    x_2 = x_space.from_str(x_str)
    if not x_space.is_equal(x, x_2):
        raise ValueError("error in space to/from_str")
    if not np.array_equal(x, x_2):
        raise ValueError("error in space to/from_str and is_equal")

    y_2 = y_space.create()
    encoding.decode(x_2, y_2)
    if not y_space.is_equal(y, y_2):
        raise ValueError("encoding is not deterministic")
    if not np.array_equal(y, y_2):
        raise ValueError(
            "encoding is not deterministic and error in space.is_equal")


def validate_signed_permutation_encoding_on_2dbinpacking(
        encoding: Encoding | Callable[[Instance], Encoding],
        random: Generator = __RANDOM) -> None:
    """
    Validate a signed permutation encoding function on bin packing instances.

    :param encoding: the encoding or a factory creating it
    :param random: the random number generator
    """
    end_time: Final[int] = monotonic_ns() + 20_000_000_000
    for i in binpacking_instances_for_tests(random):
        if monotonic_ns() >= end_time:
            break
        validate_signed_permutation_encoding_on_1_2dbinpacking(
            encoding, i, random)



"""
A class to model a dynamic system governed by differential equations.

A system has a current state vector at any point in time. The state changes
over time based on differential equations. These equations can be influenced
by the output of a :mod:`~moptipyapps.dynamic_control.controller`. Our
:class:`System` presents the state dimension and differential equations. It
also presents several starting states for simulating the system. The starting
states are divided into training and testing states. The training states can
be used for, well, training controllers to learn how to handle the system.
The testing states are then used to verify whether a synthesized controller
actually works.

Examples for different dynamic systems are given in package
:mod:`~moptipyapps.dynamic_control.systems`.
"""

from math import isfinite
from typing import Any, Callable, Final, Iterable, TypeVar

import moptipy.utils.plot_utils as pu
import numpy as np
from moptipy.api.component import Component
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import array_to_str
from pycommons.io.path import Path
from pycommons.types import check_int_range, type_error

from moptipyapps.dynamic_control.ode import multi_run_ode
from moptipyapps.dynamic_control.results_log import ResultsLog
from moptipyapps.dynamic_control.results_plot import ResultsPlot

#: the color for training points
_TRAINING_COLOR: Final[str] = "orange"
#: the color for testing points
_TEST_COLOR: Final[str] = "cyan"

#: the type variable for ODE controller parameters
T = TypeVar("T")


def _dummy_ctrl(_: np.ndarray, __: float, ___: Any,
                out: np.ndarray) -> None:
    """Do nothing as internal dummy controller that does nothing."""
    out.fill(0.0)


class System(Component):
    """A class for governing a system via differential system."""

    def __init__(self, name: str, state_dims: int, control_dims: int,
                 state_dim_mod: int, state_dims_in_j: int, gamma: float,
                 test_starting_states: np.ndarray,
                 training_starting_states: np.ndarray,
                 test_steps: int = 5000,
                 test_time: float = 50.0,
                 training_steps: int = 1000,
                 training_time: float = 50.0,
                 plot_examples: Iterable[int] = (0, )) -> None:
        """
        Initialize the system.

        :param name: the name of the system.
        :param state_dims: the state dimensions
        :param control_dims: the control dimensions
        :param state_dim_mod: the modulus for the state dimensions
        :param state_dims_in_j: the dimension until which the state is used in
            the computation of the figure of merit; `-1` for using the complete
            state
        :param gamma: the weight of the controller input
        :param test_starting_states: the starting states to be used for
            testing, as a matrix with one point per row
        :param test_steps: the steps to be taken for the test simulation
        :param test_time: the time limit for tests
        :param training_steps: the steps to be taken for the training
            simulation
        :param training_time: the time limit for training
        :param training_starting_states: the starting states to be used for
            training, as a matrix with one point per row
        :param plot_examples: the points that should be plotted
        """
        super().__init__()
        if not isinstance(name, str):
            raise type_error(name, "name", str)
        #: the name of the model
        self.name: Final[str] = name
        #: the dimensions of the state variable
        self.state_dims: Final[int] = check_int_range(
            state_dims, "state_dims", 2, 1_000_000)
        #: the dimensions of the controller output
        self.control_dims: Final[int] = check_int_range(
            control_dims, "control_dims", 1, 100)
        #: The modulus for the state dimensions for plotting
        self.state_dim_mod: Final[int] = check_int_range(
            state_dim_mod, "state_dim_mod", 0, state_dims)
        state_dims_in_j = check_int_range(
            state_dims_in_j, "state_dims_in_j", -1, state_dims)
        #: The number of dimensions used in the J computation
        self.state_dims_in_j: Final[int] = state_dims if state_dims_in_j <= 0 \
            else state_dims_in_j
        if not isinstance(gamma, float):
            raise type_error(gamma, "gamma", float)
        if (not isfinite(gamma)) and gamma > 0.0:
            raise ValueError(
                f"gamma must be positive and finite, but is {gamma}.")
        #: The Weight of the control values in the figure of merit computation.
        self.gamma: Final[float] = gamma

        if not isinstance(test_starting_states, np.ndarray):
            raise type_error(test_starting_states, "test_starting_states",
                             np.ndarray)
        if (len(test_starting_states) < 1) or (
                test_starting_states.shape[1] != state_dims):
            raise ValueError(
                f"invalid test starting states {test_starting_states!r}.")
        #: the test starting states
        self.test_starting_states: Final[np.ndarray] = test_starting_states

        if not isinstance(training_starting_states, np.ndarray):
            raise type_error(training_starting_states,
                             "training_starting_states", np.ndarray)
        if (len(training_starting_states) < 1) or (
                training_starting_states.shape[1] != state_dims):
            raise ValueError("invalid training starting "
                             f"states {training_starting_states!r}.")
        #: the test starting states
        self.training_starting_states: Final[np.ndarray] = \
            training_starting_states

        #: the test simulation steps
        self.test_steps: Final[int] = check_int_range(
            test_steps, "test_steps", 10, 1_000_000_000)

        if not isinstance(test_time, float):
            raise type_error(test_time, "test_time", float)
        if (not isfinite(test_time)) and test_time > 1e-5:
            raise ValueError(
                f"test_time must be > 1e-5 and finite, but is {test_time}.")
        #: the test time
        self.test_time: Final[float] = test_time

        #: the training simulation steps
        self.training_steps: Final[int] = check_int_range(
            training_steps, "training_steps", 10, 1_000_000_000)
        if not isinstance(training_time, float):
            raise type_error(training_time, "training_time", float)
        if (not isfinite(training_time)) and training_time > 1e-5:
            raise ValueError(f"training_time must be > 1e-5 and finite, "
                             f"but is {training_time}.")
        #: the training time
        self.training_time: Final[float] = training_time

        #: the plot examples
        self.plot_examples: Final[tuple[int, ...]] = tuple(sorted(set(
            plot_examples)))
        total: Final[int] = len(self.training_starting_states) \
            + len(self.test_starting_states) - 1
        for i in self.plot_examples:
            check_int_range(i, "plot_examples[i]", 0, total)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("stateDims", self.state_dims)
        logger.key_value("controlDims", self.control_dims)
        logger.key_value("stateDimMod", self.state_dim_mod)
        logger.key_value("stateDimsInJ", self.state_dims_in_j)
        logger.key_value("gamma", self.gamma)
        logger.key_value("testStartingStates",
                         array_to_str(self.test_starting_states.flatten()))
        logger.key_value(
            "trainingStartingStates",
            array_to_str(self.training_starting_states.flatten()))
        logger.key_value("testSteps", self.test_steps)
        logger.key_value("testTime", self.test_time)
        logger.key_value("trainingSteps", self.training_steps)
        logger.key_value("trainingTime", self.training_time)
        logger.key_value("examplePlots", array_to_str(np.array(
            self.plot_examples, int).flatten()))

    def __str__(self):
        """
        Get the name of this model.

        :return: the name of this model
        """
        return self.name

    def equations(self, state: np.ndarray, time: float,
                  control: np.ndarray, out: np.ndarray) -> None:
        """
        Compute the values of the differential equations to be simulated.

        :param state: the state of the system
        :param time: the time index
        :param control: the output of the controller
        :param out: the differential, i.e., the output of this function
        """

    def plot_points(self, dest_dir: str, skip_if_exists: bool = True) -> Path:
        """
        Plot the training and testing points of the equation.

        :param dest_dir: the destination directory
        :param skip_if_exists: `True` to skip existing files, `False` otherwise
        :return: the plotted file
        """
        name: Final[str] = f"{self.name}_points"
        dest_folder: Final[Path] = Path(dest_dir)
        dest_folder.ensure_dir_exists()
        dest_file: Final[Path] = dest_folder.resolve_inside(f"{name}.pdf")
        if dest_file.ensure_file_exists() and skip_if_exists:
            return dest_file

        figure = pu.create_figure()
        state_dims: Final[int] = self.state_dims
        state_dim_mod: Final[int] = self.state_dim_mod
        use_dims: Final[int] = state_dims if state_dim_mod <= 0 \
            else state_dim_mod
        if use_dims == 3:
            axes = figure.add_subplot(projection="3d")
            axes.set_aspect("equal")
            if hasattr(axes, "force_zorder"):
                axes.force_zorder = True
            if hasattr(axes, "set_zlabel"):
                axes.set_zlabel("z")

            if state_dims == 3:
                axes.scatter(self.training_starting_states[:, 0],
                             self.training_starting_states[:, 1],
                             self.training_starting_states[:, 2],
                             color=_TRAINING_COLOR, marker="x",
                             zorder=1)
                axes.scatter(self.test_starting_states[:, 0],
                             self.test_starting_states[:, 1],
                             self.test_starting_states[:, 2],
                             color=_TEST_COLOR, marker="o", zorder=2)
                axes.scatter(0, 0, 0, color="black", marker="+", zorder=3)
            else:
                raise NotImplementedError

            if hasattr(axes, "set_zlim"):
                if not hasattr(axes, "get_zlim"):
                    raise ValueError("Axes is invalid, does not have "
                                     "'get_zlim' but 'set_zlim'.")
                zlim = axes.get_zlim()  # type: ignore
                axes.set_zlim(1.1 * zlim[0], 1.1 * zlim[1])
        elif use_dims == 2:
            axes = pu.get_axes(figure)
            axes.set_aspect("equal")
            axes.grid(True)
            if state_dims == 2:
                axes.scatter(self.training_starting_states[:, 0],
                             self.training_starting_states[:, 1],
                             color=_TRAINING_COLOR, marker="x",
                             zorder=1)
                axes.scatter(self.test_starting_states[:, 0],
                             self.test_starting_states[:, 1],
                             color=_TEST_COLOR, marker="o", zorder=2)
            else:
                markers: Final[list[str]] = list("xvo^+<s>")
                colors: Final[list[str]] = [_TRAINING_COLOR, _TEST_COLOR]
                zorder: int = 1
                for i, data in enumerate([self.training_starting_states,
                                          self.test_starting_states]):
                    color = colors[i]
                    for row in data:
                        x: np.ndarray = row[0:state_dims:state_dim_mod]
                        y: np.ndarray = row[1:state_dims:state_dim_mod]
                        zorder += 1
                        axes.plot(x, y, color=color, zorder=zorder)
                        for k, xx in enumerate(x):
                            zorder += 1
                            axes.scatter(xx, y[k], color=color,
                                         zorder=zorder,
                                         marker=markers[k % len(markers)])
        else:
            raise ValueError(f"invalid dimensions {state_dims} "
                             f"with modulus {state_dim_mod} for {self}.")

        xlim = axes.get_xlim()
        axes.set_xlim(1.1 * xlim[0], 1.1 * xlim[1])
        ylim = axes.get_ylim()
        axes.set_ylim(1.1 * ylim[0], 1.1 * ylim[1])
        axes.set_title(f"{self.name} system\ntraining points ("
                       f"{_TRAINING_COLOR}), test points ({_TEST_COLOR})")

        axes.set_xlabel("x")
        axes.set_ylabel("y")
        res = pu.save_figure(figure, name, dest_folder, "pdf")[0]
        if res != dest_file:
            raise ValueError(
                f"Wrong destination file {res!r}!={dest_file!r}?")
        return dest_file

    def describe_system(
            self, title: str | None,
            controller: Callable[[np.ndarray, float, T, np.ndarray], None],
            parameters: T, base_name: str, dest_dir: str,
            skip_if_exists: bool = False) \
            -> tuple[Path, ...]:
        """
        Describe the performance of a given system of system.

        :param title: the title
        :param controller: the controller to simulate
        :param parameters: the controller parameters
        :param base_name: the base name of the file to produce
        :param dest_dir: the destination directory
        :param skip_if_exists: if the file already exists
        :returns: the paths of the generated files
        """
        dest: Final[Path] = Path(dest_dir)
        dest.ensure_dir_exists()

        the_title = f"{self.name} system"
        if title is not None:
            the_title = f"{the_title}: {title}"

        file_1 = dest.resolve_inside(f"{base_name}_plot.pdf")
        file_2 = dest.resolve_inside(f"{base_name}_results.csv")
        if file_1.ensure_file_exists() and file_2.ensure_file_exists() \
                and skip_if_exists:
            return file_1, file_2

        with ResultsPlot(dest_file=file_1, sup_title=the_title,
                         state_dims=self.state_dims,
                         plot_indices=self.plot_examples,
                         state_dim_mod=self.state_dim_mod) as plt, \
                ResultsLog(self.state_dims, file_2) as log:
            multi_run_ode(
                self.test_starting_states,
                self.training_starting_states,
                (plt.collector, log.collector),
                self.equations,
                controller, parameters, self.control_dims,
                self.test_steps, self.test_time,
                self.training_steps, self.training_time,
                self.state_dims_in_j, self.gamma)

        return file_1, file_2

    def describe_system_without_control(self, dest_dir: str,
                                        skip_if_exists: bool = True) \
            -> tuple[Path, ...]:
        """
        Describe the performance of a given system of system.

        :param dest_dir: the destination directory
        :param skip_if_exists: if the file already exists
        :returns: the paths of the generated files
        """
        return self.describe_system("without control", _dummy_ctrl, 0.0,
                                    f"{self.name}_without_control", dest_dir,
                                    skip_if_exists)



"""
A decoder for 2D BPP instances.

The goal of developing this decoding procedure is that we need a deterministic
mapping of some easy-to-process data structure to an instance of the
two-dimensional bin packing problem. The instance produced by the mapping
should use a pre-defined bin width, bin height, and number of items. It should
also have a pre-defined lower bound for the number of bins required and it
must be ensured that this lower bound can also be reached, i.e., that at least
one solution exists that can indeed pack all items into this number of bins.

As source data structure to be mapped, we choose the real vectors of a fixed
length (discussed later on).

The idea is that we take the bin width, bin height, and lower bound of the
number of bins (let's call it `lb`) from a template instance. We also take
the number items (let's call it `n`) from that instance.

Now we begin with `lb` items, each of which exactly of the size and dimensions
of a bin. At the end, we want to have `n` items. To get there, in each step of
our decoding, we split one existing item into two items. This means that each
step will create one additional item for the instance (while making one
existing item smaller). This, in turn, means that we have to do `n - lb`
decoding steps, as we start with `lb` items and, after `n - lb` steps, will
have `lb + n - lb = n` items.

So far so good.
But how do we split?

Each split that we want to perform be defined by four features:

1. the index of the item that we are going to split,
2. whether we split it horizontally or vertically,
3. where we are going to split it,
4. and how to continue if the proposed split is not possible, e.g., because
   it would lead to a zero-width or zero-height item.

Now we can encode this in two real numbers `selector` and `cutter` from the
interval `[-1,1]`.

First, we multiply the absolute value of the `selector` with the current
number of items that we have. This is initially `2`, will then be `3` in the
next iteration, then `4`, and so on.
Converted to an int, the result of this multiplication gives us the index of
the item to split.

Then, if `cutter >= 0`, we will cut the item horizontally. Otherwise, i.e., if
`cutter < 0`, we cut vertically.
Where to cut is then decided by multiplying the absolute value of `cutter`
with the length of the item in the selected cutting dimension.

If that is not possible, we move to the next item and try again.
If `selector < 0`, we move towards smaller indices and wrap after `0`.
Otherwise, we move towards higher indices and wrap at the end of the item
list.
If we arrive back at the first object, this means that the split was not
possible for any of the existing items.
We now rotate the split by 90°, i.e., if we tried horizontal splits, we now
try vertical ones and vice versa.

It is easy to see that it must always be possible to split at least one item
in at least one direction. Since we took the bin dimensions and numbers of
items from an existing instance of the benchmark set, it must be possible to
divide the bins into `n` items in one way or another. Therefore, the loop
will eventually terminate and yield the right amount of items.

This means that with `2 * (n - lb)` floating point numbers, we can describe an
instance whose result is a perfect packing, without any wasted space.

Now the benchmark instances are not just instances that can only be solved by
perfect packings. Instead, they have some wasted space.

We now want to add some wasted space to our instance. So far, our items occupy
exactly `lb * bin_width * bin_height` space units. We can cut at most
`bin_width * bin_height - 1` of these units without changing the required bins
of the packing: We would end up with `(lb - 1) * bin_width * bin_height + 1`
space units required by our items, which is `1` too large to fit into `lb - 1`
bins.

So we just do the same thing again:
We again use two real numbers to describe each split.
Like before, we loop through these numbers, pick the object to split, and
compute where to split it. Just now we throw away the piece that was cut off.
(Of course, we compute the split positions such that we never dip under the
area requirement discussed above).

This allows us to use additional real variables to define how the space should
be reduced. `2 * (n - lb)` variables, we get instances requiring perfect
packing. With every additional pair of variables, we cut some more space.
If we would use `2 * (n - lb) + 10` variables, then we would try to select
five items from which we can cut off a bit. This number of additional
variables can be chosen by the user.

Finally, we merge all items that have the same dimension into groups, as is
the case in some of the original instances. We then shuffle these groups
randomly, to give the instances a bit of a more unordered texture.
The random shuffling is seeded with the binary representation of the input
vectors.

In the end, we have translated a real vector to a two-dimensional bin packing
instance. Hurray.

>>> space = InstanceSpace(Instance.from_resource("a04"))
>>> print(f"{space.inst_name!r} with {space.n_different_items}/"
...       f"{space.n_items} items with area {space.total_item_area} "
...       f"in {space.min_bins} bins of "
...       f"size {space.bin_width}*{space.bin_height}.")
'a04n' with 2/16 items with area 7305688 in 3 bins of size 2750*1220.

>>> decoder = InstanceDecoder(space)
>>> import numpy as np
>>> x = np.array([ 0.0,  0.2, -0.1,  0.3,  0.5, -0.6, -0.7,  0.9,
...                0.0,  0.2, -0.1,  0.3,  0.5, -0.6, -0.7,  0.9,
...                0.0,  0.2, -0.1,  0.3,  0.5, -0.6, -0.7,  0.9,
...                0.0,  0.2, ])
>>> y = space.create()
>>> decoder.decode(x, y)
>>> space.validate(y)
>>> res: Instance = y[0]
>>> print(f"{res.name!r} with {res.n_different_items}/"
...       f"{res.n_items} items with area {res.total_item_area} "
...       f"in {res.lower_bound_bins} bins of "
...       f"size {res.bin_width}*{res.bin_height}.")
'a04n' with 15/16 items with area 10065000 in 3 bins of size 2750*1220.
>>> print(space.to_str(y))
a04n;15;2750;1220;1101,1098;2750,244;2750,98;1101,171;1649,171;2750,976;\
441,122;1649,122;2750,10;2750,1,2;2750,3;1649,1098;2750,878;2750,58;660,122


>>> x = np.array([ 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...                0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...                0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...                0.0, 0.0, ])
>>> y = space.create()
>>> decoder.decode(x, y)
>>> space.validate(y)
>>> res: Instance = y[0]
>>> print(f"{res.name!r} with {res.n_different_items}/"
...       f"{res.n_items} items with area {res.total_item_area} "
...       f"in {res.lower_bound_bins} bins of "
...       f"size {res.bin_width}*{res.bin_height}.")
'a04n' with 3/16 items with area 10065000 in 3 bins of size 2750*1220.
>>> print(space.to_str(y))
a04n;3;2750;1220;2750,1216,2;2750,1,13;2750,1215

>>> from math import nextafter
>>> a1 = nextafter(1.0, -10)
>>> x = np.array([ a1, a1, a1, a1, a1, a1, a1, a1,
...                a1, a1, a1, a1, a1, a1, a1, a1,
...                a1, a1, a1, a1, a1, a1, a1, a1,
...                a1, a1, ])
>>> y = space.create()
>>> decoder.decode(x, y)
>>> space.validate(y)
>>> res: Instance = y[0]
>>> print(f"{res.name!r} with {res.n_different_items}/"
...       f"{res.n_items} items with area {res.total_item_area} "
...       f"in {res.lower_bound_bins} bins of "
...       f"size {res.bin_width}*{res.bin_height}.")
'a04n' with 4/16 items with area 10065000 in 3 bins of size 2750*1220.
>>> print(space.to_str(y))
a04n;4;2750;1220;2750,1208;2750,1219;2750,1220;2750,1,13

>>> from math import nextafter
>>> a1 = nextafter(-1.0, 10)
>>> x = np.array([ a1, a1, a1, a1, a1, a1, a1, a1,
...                a1, a1, a1, a1, a1, a1, a1, a1,
...                a1, a1, a1, a1, a1, a1, a1, a1,
...                a1, a1, ])
>>> y = space.create()
>>> decoder.decode(x, y)
>>> space.validate(y)
>>> res: Instance = y[0]
>>> print(f"{res.name!r} with {res.n_different_items}/"
...       f"{res.n_items} items with area {res.total_item_area} "
...       f"in {res.lower_bound_bins} bins of "
...       f"size {res.bin_width}*{res.bin_height}.")
'a04n' with 5/16 items with area 10065000 in 3 bins of size 2750*1220.
>>> print(space.to_str(y))
a04n;5;2750;1220;2750,1220;2730,1220;2748,1220;1,1220,4;2,1220,9

>>> from math import nextafter
>>> a1 = nextafter(-1.0, 10)
>>> x = np.array([ a1, a1, a1, a1, a1, a1, a1, a1,
...                a1, a1, a1, a1, a1, a1, a1, a1,
...                a1, a1, a1, a1, a1, a1, a1, a1,
...                a1, a1, 0.3, 0.7 ])
>>> y = space.create()
>>> decoder.decode(x, y)
>>> space.validate(y)
>>> res: Instance = y[0]
>>> print(f"{res.name!r} with {res.n_different_items}/"
...       f"{res.n_items} items with area {res.total_item_area} "
...       f"in {res.lower_bound_bins} bins of "
...       f"size {res.bin_width}*{res.bin_height}.")
'a04n' with 6/16 items with area 10064146 in 3 bins of size 2750*1220.
>>> print(space.to_str(y))
a04n;6;2750;1220;2,1220,9;1,1220,3;2748,1220;1,366;2750,1220;2730,1220

>>> from math import nextafter
>>> a1 = nextafter(-1.0, 10)
>>> x = np.array([ a1, a1, a1, a1, a1, a1, a1, a1,
...                a1, a1, a1, a1, a1, a1, a1, a1,
...                a1, a1, a1, a1, a1, a1, a1, a1,
...                a1, a1, 0.3, 0.7, -0.2, -0.3,
...                0.5, -0.3])
>>> y = space.create()
>>> decoder.decode(x, y)
>>> space.validate(y)
>>> res: Instance = y[0]
>>> print(f"{res.name!r} with {res.n_different_items}/"
...       f"{res.n_items} items with area {res.total_item_area} "
...       f"in {res.lower_bound_bins} bins of "
...       f"size {res.bin_width}*{res.bin_height}.")
'a04n' with 6/16 items with area 10061706 in 3 bins of size 2750*1220.
>>> print(space.to_str(y))
a04n;6;2750;1220;2,1220,7;2750,1220;2730,1220;1,1220,5;1,366;2748,1220


>>> x = np.array([ 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...                0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...                0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...                0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...                0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...                0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...                0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...                0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...                0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...                0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
...                1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
...                1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
...                1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
...                1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
...                1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
...                1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
...                1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
...                -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0,
...                -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0,
...                -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0,
...                -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0,
...                -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0,
...                -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0,
...                -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0,
...                -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0,
...                -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0,
...                -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0,
...                -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0,
...                -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0,
...                -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0,
...                0.0, 0.0, ])
>>> y = space.create()
>>> decoder.decode(x, y)
>>> space.validate(y)
>>> res: Instance = y[0]
>>> print(f"{res.name!r} with {res.n_different_items}/"
...       f"{res.n_items} items with area {res.total_item_area} "
...       f"in {res.lower_bound_bins} bins of "
...       f"size {res.bin_width}*{res.bin_height}.")
'a04n' with 5/16 items with area 9910948 in 3 bins of size 2750*1220.
>>> print(space.to_str(y))
a04n;5;2750;1220;2698,1;2750,1,12;2750,1216;2750,1215;2750,1160
"""
from math import isfinite
from typing import Final

import numpy as np
from moptipy.api.encoding import Encoding
from numpy.random import default_rng
from pycommons.types import check_int_range, type_error

from moptipyapps.binpacking2d.instance import Instance
from moptipyapps.binpacking2d.instgen.instance_space import InstanceSpace


class InstanceDecoder(Encoding):
    """Decode a string of `n` real values in `[0,1]` to an instance."""

    def __init__(self, space: InstanceSpace) -> None:
        """
        Create the instance decoder.

        :param space: the instance description and space
        """
        super().__init__()
        if not isinstance(space, InstanceSpace):
            raise type_error(space, "space", InstanceSpace)
        #: the instance description
        self.space: Final[InstanceSpace] = space

    def get_x_dim(self, slack: float | int = 0) -> int:
        """
        Get the minimum dimension that a real vector must have.

        :param slack: a parameter denoting the amount of slack for reducing
            the item size
        :return: the minimum dimension
        """
        if not isinstance(slack, float | int):
            raise type_error(slack, "slack", (float, int))
        if (not isfinite(slack)) or (slack < 0):
            raise ValueError(f"slack={slack} is invalid")
        base: Final[int] = check_int_range(
            self.space.n_items - self.space.min_bins, "base", 1, 1_000_000)
        added: Final[int] = int(slack * base + 0.5)
        if (added < 0) or (added > 1_000_000):
            raise ValueError(f"invalid result {added} based "
                             f"on slack={slack} and base={base}")
        return 2 * (base + added)

    def decode(self, x: np.ndarray, y: list[Instance]) -> None:
        """
        Decode the real-valued array to a 2D BPP instance.

        :param x: an array with values in [-1, 1]
        :param y: the instance receiver
        """
        # We start by using the prescribed number of bins as items.
        # There will be n_bins items, each of which has exactly the size of a
        # bin. Therefore, we know that all items fit exactly into n_bins bins.
        bin_width: Final[int] = self.space.bin_width
        bin_height: Final[int] = self.space.bin_height
        n_bins: Final[int] = self.space.min_bins
        items: list[list[int]] = [
            [bin_width, bin_height] for _ in range(n_bins)]

        # Now we need to make n_items - n_bins cuts. Each cut will divide one
        # item into two items. Therefore, each cut yields exactly one new
        # item. Therefore, after n_items - n_bins cuts we get n_items - n_bins
        # new items. Since we start with n_bins items, this means that we get
        # n_items - n_bins + n_bins items at the end, which are exactly
        # n_items.
        #
        # Each cut cuts an item into two items along either the horizontal or
        # vertical dimension. If we would put the two new, smaller items back
        # together, they would exactly result in the item that we just cut.
        # Therefore, they do fit into the same area and their area sum is also
        # the same. Therefore, the overall area and the overall number of
        # required bins will also stay the same.
        x_idx: int = 0
        for cur_n_items in range(n_bins, self.space.n_items):
            # In each iteration of this loop, we perform one cut.
            # This means, that we add exactly one item to the item list.
            # Our counter variable `cur_n_items`, which starts at `n_bins`
            # and iterates until just before the target number of items
            # `self.desc.n_items` represents the current number of items
            # in our list `items`.
            # It thus always holds that `cur_n_items == len(items)`.

            # Each cut is described by four properties:
            # 1. The index of the item that we want to cut.
            # 2. The direction (horizontal or vertical) into which we cut.
            # 3. The location where we will cut along this dimension.
            # 4. The direction into which we will search for the next item
            #    if the current item cannot be cut this way.
            # These four properties are encoded in two real numbers `selector`
            # and `cutter` from `[-1, 1]`.

            # The first number is `sel`. It encodes (1) and (4).
            # First, we multiply `selector` with `cur_n_items` and take this
            # modulo `cur_n_items`. This gives us a value `t` with
            # `-cur_n_items < t < cur_n_items`. By adding `cur_n_items`
            # and repeating the modulo division, we get
            # `0 <= sel_i < cur_n_items`.
            # `sel_i` is the index of the item that we want to cut.
            selector: float = x[x_idx]
            x_idx += 1

            sel_i: int = ((int(cur_n_items * selector) % cur_n_items)
                          + cur_n_items) % cur_n_items
            orig_sel_i: int = sel_i  # the original selection index.

            # Now, it may not be possible to cut this item.
            # The idea is that if we cannot cut the item, we simply try to cut
            # the next item.
            # The next item could be the one at the next-higher index or the
            # one at the next lower index.
            # If `selector < 0.0`, then we will next try the item at the next
            # lower index. If `selector >= 0.0`, we will move to the next
            # higher index.
            # Of course, we will wrap our search when reaching either end of
            # the list.
            #
            # If we arrive back at `orig_sel_i`, then we have tried to cut
            # each item in the list in the prescribed cutting direction,
            # however none could be cut.
            # In this case, we will change the cutting direction.
            #
            # It must be possible to cut at least one item in at least one
            # direction, or otherwise the problem would not be solvable.
            # Therefore, we know that this way, trying all items in all
            # directions in the worst case, we will definitely succeed.
            sel_dir: int = -1 if selector < 0.0 else 1

            # `cutter` tells us where to cut and in which direction.
            # If `cutter >= 0`, then we will cut horizontally and
            # if `cutter < 0`, we cut vertically.
            # Each item is described by list `[width, height]`, so cutting
            # horizontally means picking a vertical height coordinate and
            # cutting horizontally along it. This means that for horizontal
            # cuts, the `cut_dimension` should be `1` and for vertical cuts,
            # it must be `0`.
            cutter: float = x[x_idx]
            x_idx += 1
            cut_dimension: int = 1 if cutter >= 0.0 else 0

            while True:
                cur_item: list[int] = items[sel_i]  # Get the item to cut.

                item_size_in_dim: int = cur_item[cut_dimension]

                # We define the `cut_modulus` as the modulus for the cutting
                # operation. We will definitely get a value for `cut_position`
                # such that `0 < cut_position <= cut_modulus`. This means that
                # we will get `0 < cut_position < item_size_in_dim`.
                # Therefore, if `cut_modulus > 1`, then we know that there
                # definitely is a `cut_position` at which we can cut the
                # current item and obtain two new items that have both
                # non-zero width and height.
                cut_modulus: int = item_size_in_dim - 1
                if cut_modulus > 0:  # Otherwise, we cannot cut the item.
                    cut_position: int = (((int(
                        cut_modulus * cutter) % cut_modulus) + cut_modulus)
                        % cut_modulus) + 1

                    if 0 < cut_position < item_size_in_dim:  # Sanity check...
                        # Now we perform the actual cut.
                        # The original item now gets `cut_position` as the
                        # size in the `cut_dimension`, the other one gets
                        # `item_size_in_dim - cut_position`. Therefore, the
                        # overall size remains the same.
                        cur_item[cut_dimension] = cut_position
                        cur_item = cur_item.copy()
                        cur_item[cut_dimension] = (
                            item_size_in_dim - cut_position)
                        items.append(cur_item)
                        break  # we cut one item and can stop

                sel_i = ((((sel_i + sel_dir) % cur_n_items) + cur_n_items)
                         % cur_n_items)
                if sel_i == orig_sel_i:
                    cut_dimension = 1 - cut_dimension

        # At this stage, our instance can only be solved with a perfect
        # packing. The items need to be placed perfectly together and they
        # will cover the complete `current_area`.
        bin_area: Final[int] = bin_width * bin_height
        current_area: int = n_bins * bin_area

        # However, requiring a perfect packing may add hardness to the problem
        # that does not exist in the original problem.
        # We now want to touch some of the items and make them a bit smaller.
        # However, we must never make them smaller as
        # `current_area - bin_area`, because then we could create a situation
        # where less than `n_bins` bins are required.
        # Also, we never want to slide under the `total_item_area` prescribed
        # by the original problem. If the original problem prescribed a
        # perfect packing, then we will create a perfect packing.
        min_area: Final[int] = current_area - bin_area + 1

        # If and only if the input array still has information left and if we
        # still have area that we can cut, then let's continue.
        # We will keep cutting items, but this time, we throw away the piece
        # that we cut instead of adding it as item. Of course, we never cut
        # more than what we are permitted to.
        max_x_idx: Final[int] = len(x)
        cur_n_items = list.__len__(items)
        while (x_idx < max_x_idx) and (current_area > min_area):
            # We perform the same selection and cutting choice.
            selector = x[x_idx]
            x_idx += 1
            sel_i = ((int(cur_n_items * selector) % cur_n_items)
                     + cur_n_items) % cur_n_items
            orig_sel_i = sel_i  # the original selection index.
            sel_dir = -1 if selector < 0.0 else 1
            cutter = x[x_idx]
            x_idx += 1
            cut_dimension = 1 if cutter >= 0.0 else 0

            # We have selected the item and got the cutter value, too.
            # Now we try to perform the cut. If we cannot cut, we try
            # the next item. If we could not cut any item along the
            # cut dimension, we switch the cut dimension.
            # However, we may fail to cut anything (which means that
            # we arrive at `step == 2`). In this case, we  just skip
            # this cut.
            step: int = 0
            while step < 2:  # This time, we may actually fail to cut.
                cur_item = items[sel_i]  # Get the item to cut.

                item_size_in_dim = cur_item[cut_dimension]
                item_size_in_other_dim: int = cur_item[1 - cut_dimension]
                # This time, the `cut_modulus` is also limited by the area
                # that we can actually cut. This is `current_area - min_area`.
                # Now each cut along the `cut_dimension` will cost us
                # `item_size_in_other_dim` area units.
                cut_modulus = min(
                    (current_area - min_area) // item_size_in_other_dim,
                    item_size_in_dim) - 1
                if cut_modulus > 0:
                    cut_position = (((int(
                        cut_modulus * cutter) % cut_modulus) + cut_modulus)
                        % cut_modulus) + 1

                    if 0 < cut_position < item_size_in_dim:
                        # We cut away cut_position and do not add the area at
                        # the end.
                        cur_item[cut_dimension] = \
                            item_size_in_dim - cut_position
                        break  # we cut one item and can stop

                sel_i = ((((sel_i + sel_dir) % cur_n_items) + cur_n_items)
                         % cur_n_items)
                if sel_i == orig_sel_i:
                    cut_dimension = 1 - cut_dimension
                    step += 1  # If we tried everything, this enforces a stop.

        # Finally, we sort the items in order to merge items of the
        # same dimension.
        items.sort()
        lo: int = 0
        n_items: int = list.__len__(items)
        while lo < n_items:
            # For each item in the list, we try to find all items of the same
            # dimension. Since the list is sorted, these items must all be
            # located together.
            hi: int = lo
            cur_item = items[lo]
            while (hi < n_items) and (items[hi] == cur_item):
                hi += 1
            cur_item.append(hi - lo)  # We now have the item multiplicity.
            # We delete all items with the same dimension (which must come
            # directly afterward).
            hi -= 1
            while lo < hi:
                del items[hi]
                hi -= 1
                n_items -= 1
            lo += 1  # Move to the next item

        # Now all items are sorted. This may or may not be a problem:
        # Some heuristics may either benefit or suffer if the item list
        # has a pre-defined structure. Therefore, we want to try to at least
        # somewhat remove the order and make the item list order more random.
        default_rng(int.from_bytes(x.tobytes())).shuffle(items)

        # And now we can fill in the result as the output of our encoding.
        res: Instance = Instance(  # Generate the instance
            self.space.inst_name, bin_width, bin_height, items)
        if list.__len__(y) > 0:  # If the destination has length > 0...
            y[0] = res  # ...store the instance at index 0,
        else:  # otherwise
            y.append(res)  # add the instance to it.



"""
An objective function indirectly minimizing the number of bins in packings.

This objective function computes the number of bins used. Let's call it
`n_bins`. We know the area `bin_area` of a bin as well. Now we return
`(bin_area * (n_bins - 1)) + area_of_items_in_last_bin`, where
`area_of_items_in_last_bin` is, well, the area covered by items in the
very last bin.

The idea behind this is: If one of two packings has the smaller number of
bins, then this one will always have the smaller objective value. If two
packings have the same number of bins, but one requires less space in the very
last bin, then that one is better. With this mechanism, we drive the search
towards "emptying" the last bin. If the number of items in the last bin would
reach `0`, that last bin would disappear - and we have one bin less.

This objective is similar to :mod:`~moptipyapps.binpacking2d.objectives.\
bin_count_and_small`, with the difference that it focuses on the *last* bin
whereas :mod:`~moptipyapps.binpacking2d.objectives.bin_count_and_small` tries
to minimize the area in *any* bin.
"""
from typing import Final

import numba  # type: ignore
import numpy as np
from pycommons.math.int_math import ceil_div

from moptipyapps.binpacking2d.instance import Instance
from moptipyapps.binpacking2d.objectives.bin_count import BinCount
from moptipyapps.binpacking2d.packing import (
    IDX_BIN,
    IDX_BOTTOM_Y,
    IDX_LEFT_X,
    IDX_RIGHT_X,
    IDX_TOP_Y,
)


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def bin_count_and_last_small(y: np.ndarray, bin_area: int) -> int:
    """
    Compute the number of bins and the occupied area in the last bin.

    We compute the total number of bins minus 1 and multiply it with the
    total area of items. We then add the area of items in the last bin.

    :param y: the packing
    :param bin_area: the area of a single bin
    :return: the objective value

    >>> bin_count_and_last_small(np.array([[1, 1, 10, 10, 20, 20],
    ...                                    [1, 1, 30, 30, 40, 40],
    ...                                    [1, 1, 20, 20, 30, 30]], int),
    ...                                    50*50)
    300
    >>> bin_count_and_last_small(np.array([[1, 1, 10, 10, 20, 20],
    ...                                    [1, 2, 30, 30, 40, 40],  # bin 2!
    ...                                    [1, 1, 20, 20, 30, 30]], int),
    ...                                    50*50)
    2600
    >>> bin_count_and_last_small(np.array([[1, 2, 10, 10, 20, 20],  # bin 2!
    ...                                    [1, 2, 30, 30, 40, 40],  # bin 2!
    ...                                    [1, 1, 20, 20, 30, 30]], int),
    ...                                    50*50)
    2700
    >>> bin_count_and_last_small(np.array([[1, 3, 10, 10, 20, 20],  # bin 3!
    ...                                    [1, 2, 30, 30, 40, 40],  # bin 2!
    ...                                    [1, 1, 20, 20, 30, 30]], int),
    ...                                    50*50)
    5100
    >>> bin_count_and_last_small(np.array([[1, 3, 10, 10, 50, 50],  # bin 3!
    ...                                    [1, 2, 30, 30, 60, 60],  # bin 2!
    ...                                    [1, 1, 20, 20, 30, 30]], np.int8),
    ...                                    50*50)
    6600
    """
    current_bin: int = -1  # the current idea of what the last bin is
    current_area: int = 0  # the area of items already in that bin
    n_items: Final[int] = len(y)  # the number of rows in the matrix

    for i in range(n_items):  # iterate over all packed items
        bin_idx: int = int(y[i, IDX_BIN])  # get the bin index of the item
        if bin_idx < current_bin:
            continue
        area: int = int(y[i, IDX_RIGHT_X] - y[i, IDX_LEFT_X]) \
            * int(y[i, IDX_TOP_Y] - y[i, IDX_BOTTOM_Y])
        if bin_idx > current_bin:  # it's a new biggest bin = new last bin?
            current_area = area  # then the current area is this
            current_bin = bin_idx  # and we remember it
        elif bin_idx == current_bin:  # did item go into the current last bin?
            current_area += area  # then increase size
    return (bin_area * (current_bin - 1)) + current_area  # return objective


class BinCountAndLastSmall(BinCount):
    """Compute the number of bins and the area in the last one."""

    def __init__(self, instance: Instance) -> None:
        """
        Initialize the objective function.

        :param instance: the instance to load the bounds from
        """
        super().__init__(instance)
        #: the bin size
        self._bin_size: Final[int] = instance.bin_width * instance.bin_height

    def evaluate(self, x) -> int:
        """
        Evaluate the objective function.

        :param x: the solution
        :return: the bin size and last-bin-small-area factor
        """
        return bin_count_and_last_small(x, self._bin_size)

    def to_bin_count(self, z: int) -> int:
        """
        Convert an objective value to a bin count.

        :param z: the objective value
        :return: the bin count
        """
        return ceil_div(z, self._bin_size)

    def lower_bound(self) -> int:
        """
        Get the lower bound of the number of bins and small-size objective.

        We know from the instance (:attr:`~moptipyapps.binpacking2d\
.instance.Instance.lower_bound_bins`) that we require at least as many bins
        such that they can accommodate the total area of all items together.
        Let's call this number `lb`. Now if `lb` is one, then all objects could
        be in the first bin, in which case the objective value would equal to
        the total area of all items (:attr:`~moptipyapps.binpacking2d\
.instance.Instance.total_item_area`).
        If it is `lb=2`, then we know that we will need at least two bins. The
        best case would be that almost all items are in the first bin and
        only the smallest object is in the last bin. This means that we would
        get `1 * bin_area + smallest_area` as objective value. If we have
        `lb=3` bins, then we could again have all but the smallest items
        distributed over the first two bins and only the smallest one in the
        last bin, i.e., would get `(2 * bin_area) + smallest_area`. And so on.

        :return: `total_item_area` if the lower bound `lb` of the number of
            bins is `1`, else `(lb - 1) * bin_area + smallest_area`, where
            `bin_area` is the area of a bin, `total_item_area` is the area of
            all items added up, and `smallest_area` is the area of the
            smallest item

        >>> ins = Instance("a", 100, 50, [[10, 5, 1], [3, 3, 1], [5, 5, 1]])
        >>> ins.total_item_area
        84
        >>> ins.lower_bound_bins
        1
        >>> BinCountAndLastSmall(ins).lower_bound()
        84

        >>> ins = Instance("b", 10, 50, [[10, 5, 10], [3, 3, 1], [5, 5, 1]])
        >>> ins.total_item_area
        534
        >>> ins.lower_bound_bins
        2
        >>> BinCountAndLastSmall(ins).lower_bound()
        509

        >>> ins = Instance("c", 10, 50, [[10, 5, 10], [30, 3, 1], [5, 5, 1]])
        >>> ins.total_item_area
        615
        >>> ins.lower_bound_bins
        2
        >>> BinCountAndLastSmall(ins).lower_bound()
        525
        """
        if self._instance.lower_bound_bins == 1:
            return self._instance.total_item_area
        smallest_area: int = -1
        for row in self._instance:
            area: int = int(row[0]) * int(row[1])
            if (smallest_area < 0) or (area < smallest_area):
                smallest_area = area
        return int(((self._instance.lower_bound_bins - 1)
                    * self._instance.bin_height
                    * self._instance.bin_width) + smallest_area)

    def upper_bound(self) -> int:
        """
        Get the upper bound of this objective function.

        :return: a very coarse estimate of the upper bound

        >>> ins = Instance("a", 100, 50, [[10, 5, 1], [3, 3, 1], [5, 5, 1]])
        >>> ins.n_items
        3
        >>> BinCountAndLastSmall(ins).upper_bound()
        15000

        >>> ins = Instance("b", 10, 50, [[10, 5, 10], [3, 3, 1], [5, 5, 1]])
        >>> ins.n_items
        12
        >>> BinCountAndLastSmall(ins).upper_bound()
        6000

        >>> ins = Instance("c", 10, 50, [[10, 5, 10], [30, 3, 1], [5, 5, 10]])
        >>> ins.n_items
        21
        >>> BinCountAndLastSmall(ins).upper_bound()
        10500
        """
        return self._instance.n_items * self._instance.bin_height \
            * self._instance.bin_width

    def __str__(self) -> str:
        """
        Get the name of the bins objective function.

        :return: `binCountAndLastSmall`
        :retval "binCountAndLastSmall": always
        """
        return "binCountAndLastSmall"



"""
Tools for generating 2d bin packing instances.

We want to generate instances of the two-dimensional bin packing problem.
These instances should have some pre-defined characteristics, e.g., width and
height of the bins, number of items to pack, lower bound/optimal number of
bins required by any solution, and so on.

At the same time, the instances should be hard.

We treat this whole thing as an optimization problem. Here, given are the
pre-defined instance characteristics and the goal is to find instances that
are hard to solve.

Important work on this code has been contributed by Mr. Rui ZHAO
(赵睿), <zr1329142665@163.com> a Master's student at the Institute of Applied
Optimization (应用优化研究所, http://iao.hfuu.edu.cn) of the School of
Artificial Intelligence and Big Data (人工智能与大数据学院) at Hefei University
(合肥大学) in Hefei, Anhui, China (中国安徽省合肥市) under the supervision of
Prof. Dr. Thomas Weise (汤卫思教授).
"""



"""An extended end result statistics record to represent packings."""
import argparse
import os.path
from dataclasses import dataclass
from math import isfinite
from typing import Any, Callable, Final, Iterable, Mapping, cast

from moptipy.evaluation.base import (
    KEY_N,
    EvaluationDataElement,
)
from moptipy.evaluation.end_results import EndResult
from moptipy.evaluation.end_statistics import CsvReader as EsCsvReader
from moptipy.evaluation.end_statistics import CsvWriter as EsCsvWriter
from moptipy.evaluation.end_statistics import (
    EndStatistics,
)
from moptipy.evaluation.end_statistics import (
    from_end_results as es_from_end_results,
)
from moptipy.utils.strings import (
    num_to_str,
)
from pycommons.ds.immutable_map import immutable_mapping
from pycommons.ds.sequences import reiterable
from pycommons.io.console import logger
from pycommons.io.csv import (
    SCOPE_SEPARATOR,
    csv_column,
    csv_read,
    csv_scope,
    csv_select_scope,
    csv_write,
)
from pycommons.io.path import Path, file_path, line_writer
from pycommons.math.sample_statistics import CsvReader as SsCsvReader
from pycommons.math.sample_statistics import CsvWriter as SsCsvWriter
from pycommons.math.sample_statistics import SampleStatistics
from pycommons.math.sample_statistics import from_samples as ss_from_samples
from pycommons.strings.string_conv import str_to_num
from pycommons.types import check_int_range, type_error

from moptipyapps.binpacking2d.objectives.bin_count import BIN_COUNT_NAME
from moptipyapps.binpacking2d.packing_result import (
    _OBJECTIVE_LOWER,
    _OBJECTIVE_UPPER,
    KEY_BIN_HEIGHT,
    KEY_BIN_WIDTH,
    KEY_N_DIFFERENT_ITEMS,
    KEY_N_ITEMS,
    LOWER_BOUNDS_BIN_COUNT,
    PackingResult,
)
from moptipyapps.binpacking2d.packing_result import from_csv as pr_from_csv
from moptipyapps.binpacking2d.packing_result import from_logs as pr_from_logs
from moptipyapps.shared import (
    moptipyapps_argparser,
    motipyapps_footer_bottom_comments,
)


@dataclass(frozen=True, init=False, order=False, eq=False)
class PackingStatistics(EvaluationDataElement):
    """
    An end statistics record of one run of one algorithm on one problem.

    This record provides the information of the outcome of one application of
    one algorithm to one problem instance in an immutable way.
    """

    #: the original end statistics record
    end_statistics: EndStatistics
    #: the number of items in the instance
    n_items: int
    #: the number of different items in the instance
    n_different_items: int
    #: the bin width
    bin_width: int
    #: the bin height
    bin_height: int
    #: the objective values evaluated after the optimization
    objectives: Mapping[str, SampleStatistics]
    #: the bounds for the objective values (append ".lowerBound" and
    #: ".upperBound" to all objective function names)
    objective_bounds: Mapping[str, int | float]
    #: the bounds for the minimum number of bins of the instance
    bin_bounds: Mapping[str, int]

    def __init__(self,
                 end_statistics: EndStatistics,
                 n_items: int,
                 n_different_items: int,
                 bin_width: int,
                 bin_height: int,
                 objectives: Mapping[str, SampleStatistics],
                 objective_bounds: Mapping[str, int | float],
                 bin_bounds: Mapping[str, int]):
        """
        Create a consistent instance of :class:`PackingStatistics`.

        :param end_statistics: the end statistics
        :param n_items: the number of items
        :param n_different_items: the number of different items
        :param bin_width: the bin width
        :param bin_height: the bin height
        :param objectives: the objective values computed after the
            optimization
        :param bin_bounds: the different bounds for the number of bins
        :param objective_bounds: the bounds for the objective functions
        :raises TypeError: if any parameter has a wrong type
        :raises ValueError: if the parameter values are inconsistent
        """
        super().__init__()
        if not isinstance(end_statistics, EndStatistics):
            raise type_error(end_statistics, "end_statistics", EndResult)
        if end_statistics.best_f != objectives[end_statistics.objective]:
            raise ValueError(
                f"end_statistics.best_f={end_statistics.best_f}, but "
                f"objectives[{end_statistics.objective!r}]="
                f"{objectives[end_statistics.objective]}.")
        if not isinstance(objectives, Mapping):
            raise type_error(objectives, "objectives", Mapping)
        if not isinstance(objective_bounds, Mapping):
            raise type_error(objective_bounds, "objective_bounds", Mapping)
        if not isinstance(bin_bounds, Mapping):
            raise type_error(bin_bounds, "bin_bounds", Mapping)
        if len(objective_bounds) != (2 * len(objectives)):
            raise ValueError(f"it is required that there is a lower and an "
                             f"upper bound for each of the {len(objectives)} "
                             f"functions, but we got {len(objective_bounds)} "
                             f"bounds, objectives={objectives}, "
                             f"objective_bounds={objective_bounds}.")

        for name, stat in objectives.items():
            if not isinstance(name, str):
                raise type_error(
                    name, f"name of evaluation[{name!r}]={stat!r}", str)
            if not isinstance(stat, int | float | SampleStatistics):
                raise type_error(
                    stat, f"value of evaluation[{name!r}]={stat!r}",
                    (int, float, SampleStatistics))
            lll: str = csv_scope(name, _OBJECTIVE_LOWER)
            lower = objective_bounds[lll]
            if not isfinite(lower):
                raise ValueError(f"{lll}=={lower}.")
            uuu = csv_scope(name, _OBJECTIVE_UPPER)
            upper = objective_bounds[uuu]
            for value in (stat.minimum, stat.maximum) \
                    if isinstance(stat, SampleStatistics) else (stat, ):
                if not isfinite(value):
                    raise ValueError(
                        f"non-finite value of evaluation[{name!r}]={value!r}")
                if not (lower <= value <= upper):
                    raise ValueError(
                        f"it is required that {lll}<=f<={uuu}, but got "
                        f"{lower}, {value}, and {upper}.")
        bins: Final[SampleStatistics | None] = cast(
            SampleStatistics, objectives[BIN_COUNT_NAME]) \
            if BIN_COUNT_NAME in objectives else None
        for name2, value2 in bin_bounds.items():
            if not isinstance(name2, str):
                raise type_error(
                    name2, f"name of bounds[{name2!r}]={value2!r}", str)
            check_int_range(value2, f"bounds[{name2!r}]", 1, 1_000_000_000)
            if (bins is not None) and (bins.minimum < value2):
                raise ValueError(
                    f"number of bins={bins} is inconsistent with "
                    f"bound {name2!r}={value2}.")

        object.__setattr__(self, "end_statistics", end_statistics)
        object.__setattr__(self, "objectives", immutable_mapping(objectives))
        object.__setattr__(self, "objective_bounds",
                           immutable_mapping(objective_bounds))
        object.__setattr__(self, "bin_bounds", immutable_mapping(bin_bounds))
        object.__setattr__(self, "n_different_items", check_int_range(
            n_different_items, "n_different_items", 1, 1_000_000_000_000))
        object.__setattr__(self, "n_items", check_int_range(
            n_items, "n_items", n_different_items, 1_000_000_000_000))
        object.__setattr__(self, "bin_width", check_int_range(
            bin_width, "bin_width", 1, 1_000_000_000_000))
        object.__setattr__(self, "bin_height", check_int_range(
            bin_height, "bin_height", 1, 1_000_000_000_000))

    def _tuple(self) -> tuple[Any, ...]:
        """
        Create a tuple with all the data of this data class for comparison.

        :returns: a tuple with all the data of this class, where `None` values
            are masked out
        """
        # noinspection PyProtectedMember
        return self.end_statistics._tuple()


def from_packing_results(
        results: Iterable[PackingResult],
        consumer: Callable[[PackingStatistics], None]) -> None:
    """
    Create packing statistics from a sequence of packing results.

    :param results: the packing results
    :param consumer: the consumer receiving the created packing
        statistics
    """
    if not isinstance(results, Iterable):
        raise type_error(results, "results", Iterable)
    if not callable(consumer):
        raise type_error(consumer, "consumer", call=True)
    groups: Final[dict[tuple[str, str, str, str], list[PackingResult]]] \
        = {}
    objectives_set: set[str] = set()
    for i, pr in enumerate(results):
        if not isinstance(pr, PackingResult):
            raise type_error(pr, f"end_results[{i}]", PackingResult)
        setting: tuple[str, str, str, str] = \
            (pr.end_result.algorithm, pr.end_result.instance,
             pr.end_result.objective, "" if pr.end_result.encoding is None
             else pr.end_result.encoding)
        if setting in groups:
            groups[setting].append(pr)
        else:
            groups[setting] = [pr]
        objectives_set.update(pr.objectives.keys())

    if len(groups) <= 0:
        raise ValueError("results is empty!")
    if len(objectives_set) <= 0:
        raise ValueError("results has not objectives!")
    end_stats: Final[list[EndStatistics]] = []
    objectives: Final[list[str]] = sorted(objectives_set)

    for key in sorted(groups.keys()):
        data = groups[key]
        pr0 = data[0]
        n_items: int = pr0.n_items
        n_different_items: int = pr0.n_different_items
        bin_width: int = pr0.bin_width
        bin_height: int = pr0.bin_height
        used_objective: str = pr0.end_result.objective
        encoding: str | None = pr0.end_result.encoding
        if used_objective not in objectives_set:
            raise ValueError(
                f"{used_objective!r} not in {objectives_set!r}.")
        if used_objective != key[2]:
            raise ValueError(
                f"used objective={used_objective!r} different "
                f"from key[2]={key[2]}!?")
        if (encoding is not None) and (encoding != key[3]):
            raise ValueError(
                f"used encoding={encoding!r} different "
                f"from key[3]={key[3]}!?")
        objective_bounds: Mapping[str, int | float] = pr0.objective_bounds
        bin_bounds: Mapping[str, int] = pr0.bin_bounds
        for i, pr in enumerate(data):
            if n_items != pr.n_items:
                raise ValueError(f"n_items={n_items} for data[0] but "
                                 f"{pr.n_items} for data[{i}]?")
            if n_different_items != pr.n_different_items:
                raise ValueError(
                    f"n_different_items={n_different_items} for data[0] "
                    f"but {pr.n_different_items} for data[{i}]?")
            if bin_width != pr.bin_width:
                raise ValueError(
                    f"bin_width={bin_width} for data[0] "
                    f"but {pr.bin_width} for data[{i}]?")
            if bin_height != pr.bin_height:
                raise ValueError(
                    f"bin_height={bin_height} for data[0] "
                    f"but {pr.bin_height} for data[{i}]?")
            if used_objective != pr.end_result.objective:
                raise ValueError(
                    f"used objective={used_objective!r} for data[0] "
                    f"but {pr.end_result.objective!r} for data[{i}]?")
            if objective_bounds != pr.objective_bounds:
                raise ValueError(
                    f"objective_bounds={objective_bounds!r} for data[0] "
                    f"but {pr.objective_bounds!r} for data[{i}]?")
            if bin_bounds != pr.bin_bounds:
                raise ValueError(
                    f"bin_bounds={bin_bounds!r} for data[0] "
                    f"but {pr.bin_bounds!r} for data[{i}]?")

        es_from_end_results((pr.end_result for pr in data), end_stats.append)
        if len(end_stats) != 1:
            raise ValueError(f"got {end_stats} from {data}?")

        consumer(PackingStatistics(
            end_statistics=end_stats[0],
            n_items=n_items,
            n_different_items=n_different_items,
            bin_width=bin_width,
            bin_height=bin_height,
            objectives={
                o: ss_from_samples(pr.objectives[o] for pr in data)
                for o in objectives
            },
            objective_bounds=objective_bounds,
            bin_bounds=bin_bounds,
        ))
        end_stats.clear()


def to_csv(results: Iterable[PackingStatistics], file: str) -> Path:
    """
    Write a sequence of packing statistics to a file in CSV format.

    :param results: the end statistics
    :param file: the path
    :return: the path of the file that was written
    """
    path: Final[Path] = Path(file)
    logger(f"Writing packing statistics to CSV file {path!r}.")
    path.ensure_parent_dir_exists()
    with path.open_for_write() as wt:
        csv_write(
            data=sorted(results), consumer=line_writer(wt),
            setup=CsvWriter().setup,
            get_column_titles=CsvWriter.get_column_titles,
            get_row=CsvWriter.get_row,
            get_header_comments=CsvWriter.get_header_comments,
            get_footer_comments=CsvWriter.get_footer_comments,
            get_footer_bottom_comments=CsvWriter.get_footer_bottom_comments)
    logger(f"Done writing packing statistics to CSV file {path!r}.")
    return path


def from_csv(file: str,
             consumer: Callable[[PackingStatistics], None]) -> None:
    """
    Load the packing statistics from a CSV file.

    :param file: the file to read from
    :param consumer: the consumer for the statistics
    """
    if not callable(consumer):
        raise type_error(consumer, "consumer", call=True)
    path: Final[Path] = file_path(file)
    logger(f"Now reading CSV file {path!r}.")
    with path.open_for_read() as rd:
        csv_read(rows=rd,
                 setup=CsvReader,
                 parse_row=CsvReader.parse_row,
                 consumer=consumer)
    logger(f"Done reading CSV file {path!r}.")


class CsvWriter:
    """A class for CSV writing of :class:`PackingStatistics`."""

    def __init__(self, scope: str | None = None) -> None:
        """
        Initialize the csv writer.

        :param scope: the prefix to be pre-pended to all columns
        """
        #: an optional scope
        self.scope: Final[str | None] = (
            str.strip(scope)) if scope is not None else None

        #: has this writer been set up?
        self.__setup: bool = False
        #: the end statistics writer
        self.__es: Final[EsCsvWriter] = EsCsvWriter(scope)
        #: the bin bounds
        self.__bin_bounds: list[str] | None = None
        #: the objectives
        self.__objectives: list[SsCsvWriter] | None = None
        #: the objective names
        self.__objective_names: tuple[str, ...] | None = None
        #: the lower bound names
        self.__objective_lb_names: tuple[str, ...] | None = None
        #: the upper bound names
        self.__objective_ub_names: tuple[str, ...] | None = None

    def setup(self, data: Iterable[PackingStatistics]) -> "CsvWriter":
        """
        Set up this csv writer based on existing data.

        :param data: the data to setup with
        :returns: this writer
        """
        if self.__setup:
            raise ValueError("CSV writer has already been set up.")
        self.__setup = True

        data = reiterable(data)
        self.__es.setup(pr.end_statistics for pr in data)

        bin_bounds_set: Final[set[str]] = set()
        objectives_set: Final[set[str]] = set()
        for pr in data:
            bin_bounds_set.update(pr.bin_bounds.keys())
            objectives_set.update(pr.objectives.keys())
        if set.__len__(bin_bounds_set) > 0:
            self.__bin_bounds = sorted(bin_bounds_set)
        if set.__len__(objectives_set) > 0:
            p: Final[str | None] = self.scope
            self.__objective_names = tuple(sorted(objectives_set))
            self.__objective_lb_names = tuple(csv_scope(
                oxx, _OBJECTIVE_LOWER) for oxx in self.__objective_names)
            self.__objective_ub_names = tuple(csv_scope(
                oxx, _OBJECTIVE_UPPER) for oxx in self.__objective_names)
            self.__objectives = [SsCsvWriter(
                scope=csv_scope(p, k), n_not_needed=True, what_short=k,
                what_long=f"objective function {k}").setup(
                    ddd.objectives[k] for ddd in data
            ) for k in self.__objective_names]

        return self

    def get_column_titles(self, dest: Callable[[str], None]) -> None:
        """
        Get the column titles.

        :param dest: the destination string consumer
        """
        p: Final[str | None] = self.scope
        self.__es.get_column_titles(dest)

        dest(csv_scope(p, KEY_BIN_HEIGHT))
        dest(csv_scope(p, KEY_BIN_WIDTH))
        dest(csv_scope(p, KEY_N_ITEMS))
        dest(csv_scope(p, KEY_N_DIFFERENT_ITEMS))
        if self.__bin_bounds:
            for b in self.__bin_bounds:
                dest(csv_scope(p, b))
        if self.__objective_names and self.__objectives:
            for i, o in enumerate(self.__objectives):
                dest(csv_scope(p, self.__objective_lb_names[i]))
                o.get_column_titles(dest)
                dest(csv_scope(p, self.__objective_ub_names[i]))

    def get_row(self, data: PackingStatistics,
                dest: Callable[[str], None]) -> None:
        """
        Render a single packing result record to a CSV row.

        :param data: the end result record
        :param dest: the string consumer
        """
        self.__es.get_row(data.end_statistics, dest)
        dest(repr(data.bin_height))
        dest(repr(data.bin_width))
        dest(repr(data.n_items))
        dest(repr(data.n_different_items))
        if self.__bin_bounds:
            for bb in self.__bin_bounds:
                dest(repr(data.bin_bounds[bb])
                     if bb in data.bin_bounds else "")
        if self.__objective_names and self.__objectives:
            lb: Final[tuple[str, ...] | None] = self.__objective_lb_names
            ub: Final[tuple[str, ...] | None] = self.__objective_ub_names
            for i, ob in enumerate(self.__objective_names):
                if lb is not None:
                    ox = lb[i]
                    dest(num_to_str(data.objective_bounds[ox])
                         if ox in data.objective_bounds else "")
                SsCsvWriter.get_optional_row(
                    self.__objectives[i], data.objectives.get(ob), dest)
                if ub is not None:
                    ox = ub[i]
                    dest(num_to_str(data.objective_bounds[ox])
                         if ox in data.objective_bounds else "")

    def get_header_comments(self, dest: Callable[[str], None]) -> None:
        """
        Get any possible header comments.

        :param dest: the destination
        """
        dest("End Statistics of Bin Packing Experiments")
        dest("See the description at the bottom of the file.")

    def get_footer_comments(self, dest: Callable[[str], None]) -> None:
        """
        Get any possible footer comments.

        :param dest: the destination
        """
        self.__es.get_footer_comments(dest)
        dest("")
        p: Final[str | None] = self.scope
        if self.__bin_bounds:
            for bb in self.__bin_bounds:
                dest(f"{csv_scope(p, bb)} is a lower "
                     "bound for the number of bins.")
        if self.__objectives and self.__objective_names:
            for i, obb in enumerate(self.__objective_names):
                ob: str = csv_scope(p, obb)
                ox: str = csv_scope(ob, _OBJECTIVE_LOWER)
                dest(f"{ox}: a lower bound of the {ob} objective function.")
                self.__objectives[i].get_footer_comments(dest)
                ox = csv_scope(ob, _OBJECTIVE_UPPER)
                dest(f"{ox}: an upper bound of the {ob} objective function.")

    def get_footer_bottom_comments(self, dest: Callable[[str], None]) -> None:
        """
        Get the bottom footer comments.

        :param dest: the destination
        """
        motipyapps_footer_bottom_comments(
            self, dest, "The packing data is assembled using module "
                        "moptipyapps.binpacking2d.packing_statistics.")
        EsCsvWriter.get_footer_bottom_comments(self.__es, dest)


class CsvReader:
    """A class for CSV parsing to get :class:`PackingStatistics`."""

    def __init__(self, columns: dict[str, int]) -> None:
        """
        Create a CSV parser for :class:`EndResult`.

        :param columns: the columns
        """
        super().__init__()
        #: the end result csv reader
        self.__es: Final[EsCsvReader] = EsCsvReader(columns)
        #: the index of the n-items column
        self.__idx_n_items: Final[int] = csv_column(columns, KEY_N_ITEMS)
        #: the index of the n different items column
        self.__idx_n_different: Final[int] = csv_column(
            columns, KEY_N_DIFFERENT_ITEMS)
        #: the index of the bin width column
        self.__idx_bin_width: Final[int] = csv_column(
            columns, KEY_BIN_WIDTH)
        #: the index of the bin height column
        self.__idx_bin_height: Final[int] = csv_column(
            columns, KEY_BIN_HEIGHT)
        #: the indices for the objective bounds
        self.__bin_bounds: Final[tuple[tuple[str, int], ...]] = \
            csv_select_scope(
                lambda x: tuple(sorted(((k, v) for k, v in x.items()))),
                columns, LOWER_BOUNDS_BIN_COUNT)
        if tuple.__len__(self.__bin_bounds) <= 0:
            raise ValueError("No bin bounds found?")
        #: the objective bounds columns
        self.__objective_bounds: Final[tuple[tuple[str, int], ...]] = \
            csv_select_scope(
                lambda x: tuple(sorted(((k, v) for k, v in x.items()))),
                columns, None,
                skip_orig_key=lambda s: not str.endswith(
                    s, (_OBJECTIVE_LOWER, _OBJECTIVE_UPPER)))
        n_bounds: Final[int] = tuple.__len__(self.__objective_bounds)
        if n_bounds <= 0:
            raise ValueError("No objective function bounds found?")
        if (n_bounds & 1) != 0:
            raise ValueError(f"Number of bounds {n_bounds} should be even.")
        n_val: Final[tuple[tuple[str, int]]] = ((KEY_N, self.__es.idx_n), )
        #: the parsers for the per-objective statistics
        self.__objectives: Final[tuple[tuple[str, SsCsvReader], ...]] = \
            tuple((ss, csv_select_scope(SsCsvReader, columns, ss, n_val))
                  for ss in sorted({s[0] for s in (str.split(
                      kk[0], SCOPE_SEPARATOR) for kk in
                      self.__objective_bounds) if (list.__len__(s) > 1)
                      and (str.__len__(s[0]) > 0)}))
        n_objectives: Final[int] = tuple.__len__(self.__objectives)
        if n_objectives <= 0:
            raise ValueError("No objectives found?")
        if (2 * n_objectives) != n_bounds:
            raise ValueError(
                f"Number {n_objectives} of objectives "
                f"inconsistent with number {n_bounds} of bounds.")

    def parse_row(self, data: list[str]) -> PackingStatistics:
        """
        Parse a row of data.

        :param data: the data row
        :return: the end result statistics
        """
        return PackingStatistics(
            self.__es.parse_row(data),
            int(data[self.__idx_n_items]),
            int(data[self.__idx_n_different]),
            int(data[self.__idx_bin_width]),
            int(data[self.__idx_bin_height]),
            {o: v.parse_row(data) for o, v in self.__objectives},
            {o: str_to_num(data[v]) for o, v in self.__objective_bounds},
            {o: int(data[v]) for o, v in self.__bin_bounds},
        )


# Run packing-results to stat file if executed as script
if __name__ == "__main__":
    parser: Final[argparse.ArgumentParser] = moptipyapps_argparser(
        __file__, "Build an end-results statistics CSV file.",
        "This program computes statistics over packing results")
    def_src: str = "./evaluation/end_results.txt"
    if not os.path.isfile(def_src):
        def_src = "./results"
    parser.add_argument(
        "source", nargs="?", default=def_src,
        help="either the directory with moptipy log files or the path to the "
             "end-results CSV file", type=Path)
    parser.add_argument(
        "dest", type=Path, nargs="?",
        default="./evaluation/end_statistics.txt",
        help="the path to the end results statistics CSV file to be created")
    args: Final[argparse.Namespace] = parser.parse_args()

    src_path: Final[Path] = args.source
    packing_results: Final[list[PackingResult]] = []
    if src_path.is_file():
        logger(f"{src_path!r} identifies as file, load as end-results csv")
        pr_from_csv(src_path, packing_results.append)
    else:
        logger(f"{src_path!r} identifies as directory, load it as log files")
        pr_from_logs(src_path, packing_results.append)

    packing_stats: Final[list[PackingStatistics]] = []
    from_packing_results(
        results=packing_results, consumer=packing_stats.append)
    to_csv(packing_stats, args.dest)



"""
An objective function assessing the hardness of an instance.

>>> from moptipyapps.binpacking2d.instgen.instance_space import InstanceSpace
>>> orig = Instance.from_resource("a04")
>>> space = InstanceSpace(orig)
>>> print(f"{space.inst_name!r} with {space.n_different_items}/"
...       f"{space.n_items} items with area {space.total_item_area} "
...       f"in {space.min_bins} bins of "
...       f"size {space.bin_width}*{space.bin_height}.")
'a04n' with 2/16 items with area 7305688 in 3 bins of size 2750*1220.

>>> from moptipyapps.binpacking2d.instgen.inst_decoding import InstanceDecoder
>>> decoder = InstanceDecoder(space)
>>> import numpy as np
>>> x = np.array([ 0.0,  0.2, -0.1,  0.3,  0.5, -0.6, -0.7,  0.9,
...                0.0,  0.2, -0.1,  0.3,  0.5, -0.6, -0.7,  0.9,
...                0.0,  0.2, -0.1,  0.3,  0.5, -0.6, -0.7,  0.9,
...                0.0,  0.2, ])
>>> y = space.create()
>>> decoder.decode(x, y)
>>> space.validate(y)
>>> res: Instance = y[0]
>>> print(f"{res.name!r} with {res.n_different_items}/"
...       f"{res.n_items} items with area {res.total_item_area} "
...       f"in {res.lower_bound_bins} bins of "
...       f"size {res.bin_width}*{res.bin_height}.")
'a04n' with 15/16 items with area 10065000 in 3 bins of size 2750*1220.
>>> print(space.to_str(y))
a04n;15;2750;1220;1101,1098;2750,244;2750,98;1101,171;1649,171;2750,976;\
441,122;1649,122;2750,10;2750,1,2;2750,3;1649,1098;2750,878;2750,58;660,122

>>> hardness = Hardness(max_fes=100)
>>> hardness.lower_bound()
0.0
>>> hardness.upper_bound()
1.0
>>> hardness.evaluate(y)
0.4916909965018594

>>> y[0] = orig
>>> hardness.evaluate(y)
0.9606359978301315

>>> z = Instance.from_compact_str(
...     "cl04_020_01n;19;100;100;1,10;2,38;2,62;1,4,2;3,38;1,7;27,93;1,62;1,"
...     "3;13,38;1,38;1,17;1,45;36,62;39,3;1,2;20,10;3,24;12,4")
>>> hardness.evaluate(z)
0.6430950952328174
"""
from math import isfinite
from typing import Callable, Final, Iterable

from moptipy.algorithms.random_sampling import RandomSampling
from moptipy.algorithms.so.rls import RLS
from moptipy.api.execution import Execution
from moptipy.api.objective import Objective
from moptipy.operators.signed_permutations.op0_shuffle_and_flip import (
    Op0ShuffleAndFlip,
)
from moptipy.operators.signed_permutations.op1_swap_2_or_flip import (
    Op1Swap2OrFlip,
)
from moptipy.spaces.signed_permutations import SignedPermutations
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import rand_seeds_from_str
from pycommons.types import check_int_range

from moptipyapps.binpacking2d.encodings.ibl_encoding_1 import (
    ImprovedBottomLeftEncoding1,
)
from moptipyapps.binpacking2d.instance import (
    Instance,
)
from moptipyapps.binpacking2d.objectives.bin_count import BinCount
from moptipyapps.binpacking2d.objectives.bin_count_and_last_skyline import (
    BinCountAndLastSkyline,
)
from moptipyapps.binpacking2d.packing_space import PackingSpace


def setup_rls_f7(instance: Instance) -> tuple[Execution, Objective]:
    """
    Set up the randomized local search for an instance.

    :param instance: the instance
    :return: the execution and upper bound of the objective
    """
    search_space = SignedPermutations(
        instance.get_standard_item_sequence())  # Create the search space.
    solution_space = PackingSpace(instance)  # Create the space of packings.
    objective: Final[BinCountAndLastSkyline] = BinCountAndLastSkyline(instance)

    # Build a single execution of a single run of a single algorithm and
    # return the upper bound of the objective value
    return (Execution()
            .set_search_space(search_space)
            .set_solution_space(solution_space)
            .set_encoding(ImprovedBottomLeftEncoding1(instance))
            .set_algorithm(  # This is the algorithm: Randomized Local Search.
                RLS(Op0ShuffleAndFlip(search_space), Op1Swap2OrFlip()))
            .set_objective(objective), objective)


def setup_rls_f1(instance: Instance) -> tuple[Execution, Objective]:
    """
    Set up the randomized local search for an instance.

    :param instance: the instance
    :return: the execution
    """
    search_space = SignedPermutations(
        instance.get_standard_item_sequence())  # Create the search space.
    solution_space = PackingSpace(instance)  # Create the space of packings.
    objective: Final[BinCount] = BinCount(instance)

    # Build a single execution of a single run of a single algorithm and
    # return the lower bound of the objective value
    return (Execution()
            .set_search_space(search_space)
            .set_solution_space(solution_space)
            .set_encoding(ImprovedBottomLeftEncoding1(instance))
            .set_algorithm(  # This is the algorithm: Randomized Local Search.
                RLS(Op0ShuffleAndFlip(search_space), Op1Swap2OrFlip()))
            .set_objective(objective), objective)


def setup_rs_f1(instance: Instance) -> tuple[Execution, Objective]:
    """
    Set up the randomized sampling for an instance.

    :param instance: the instance
    :return: the execution
    """
    search_space = SignedPermutations(
        instance.get_standard_item_sequence())  # Create the search space.
    solution_space = PackingSpace(instance)  # Create the space of packings.
    objective: Final[BinCount] = BinCount(instance)

    # Build a single execution of a single run of a single algorithm and
    # return the lower bound of the objective value
    return (Execution()
            .set_search_space(search_space)
            .set_solution_space(solution_space)
            .set_encoding(ImprovedBottomLeftEncoding1(instance))
            .set_algorithm(  # This is the algorithm: Random Sampling
                RandomSampling(Op0ShuffleAndFlip(search_space)))
            .set_objective(objective), objective)


#: the default executors
DEFAULT_EXECUTORS: Final[tuple[Callable[[Instance], tuple[
    Execution, Objective]], ...]] = (
        setup_rs_f1, setup_rls_f1, setup_rls_f7)


class Hardness(Objective):
    """Compute the hardness of an instance."""

    def __init__(
            self, max_fes: int = 1_000_000, n_runs: int = 3,
            executors: Iterable[Callable[[Instance], tuple[
                Execution, Objective]]] = DEFAULT_EXECUTORS) -> None:
        """
        Initialize the hardness objective function.

        :param max_fes: the maximum FEs
        :param n_runs: the maximum runs
        :param executors: the functions creating the executions
        """
        super().__init__()
        #: the maximum FEs per setup.
        self.max_fes: Final[int] = check_int_range(
            max_fes, "max_fes", 2, 1_000_000_000_000)
        #: the maximum FEs per setup.
        self.n_runs: Final[int] = check_int_range(
            n_runs, "n_runs", 1, 1_000_000)
        #: the executors
        self.executors: Final[tuple[Callable[[Instance], tuple[
            Execution, Objective]], ...]] = tuple(executors)

        #: the last instance name
        self.__last_inst: str | None = None
        #: the last seeds name
        self.__last_seeds: tuple[int, ...] | None = None

    def evaluate(self, x: list[Instance] | Instance) -> float:
        """
        Compute the hardness of an instance.

        :param x: the instance
        :return: the hardness
        """
        instance: Final[Instance] = x[0] if isinstance(x, list) else x
        seeds: tuple[int, ...]

        name: str = instance.name
        if (self.__last_seeds is None) or (self.__last_inst is None) or (
                self.__last_inst != name):
            self.__last_seeds = seeds = tuple(rand_seeds_from_str(
                f"seed for {instance.name}", self.n_runs))
            self.__last_inst = name
        else:
            seeds = self.__last_seeds

        max_fes: Final[int] = self.max_fes
        runs: int = 0
        result: float = 0.0
        for executor in self.executors:
            execs, f = executor(instance)
            lb: int | float = f.lower_bound()
            ub: int | float = f.upper_bound()
            if not (isfinite(lb) and isfinite(ub) and (lb < ub)):
                raise ValueError(f"Invalid lower and upper bound {lb}, {ub}.")
            execs.set_max_fes(max_fes)
            for seed in seeds:
                execs.set_rand_seed(seed)
                with execs.execute() as proc:
                    runs += 1
                    quality: int | float = proc.get_best_f()
                    if not (isfinite(quality) and (lb <= quality <= ub)):
                        raise ValueError(
                            f"quality={quality} invalid, must be in "
                            f"[{lb}, {ub}] for objective {f}.")
                    quality = (ub - quality) / (ub - lb)
                    if not (0.0 <= quality <= 1.0):
                        raise ValueError(
                            f"invalid normalized quality {quality} "
                            f"for objective {f}.")
                    runtime: int | float = proc.get_last_improvement_fe()
                    if not (0 < runtime <= max_fes):
                        raise ValueError(f"invalid FEs {runtime}, must "
                                         f"be in 1..{max_fes}.")
                    runtime = (max_fes - runtime) / (max_fes - 1)
                    if not (0.0 <= runtime <= 1.0):
                        raise ValueError(
                            f"invalid normalized runtime {runtime}.")
                    result += quality * runtime
        return result / runs

    def lower_bound(self) -> float:
        """
        Get the lower bound of the instance hardness.

        :return: the lower bound for the instance hardness
        :returns 0.0: always
        """
        return 0.0

    def upper_bound(self) -> float:
        """
        Get the upper bound of the instance hardness.

        :return: the upper bound for the instance hardness
        :returns 1.0: always
        """
        return 1.0

    def is_always_integer(self) -> bool:
        """
        Return `False` because the hardness function returns `float`.

        :retval False: always
        """
        return False

    def __str__(self) -> str:
        """
        Get the name of the hardness objective function.

        :return: `hardness`
        :retval "hardness": always
        """
        return "hardness"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this instance.

        :param logger: the logger
        """
        super().log_parameters_to(logger)
        logger.key_value("nRuns", self.n_runs)
        logger.key_value("maxFEs", self.max_fes)
        logger.key_value("nExecutors", len(self.executors))



"""
Here we provide a :class:`~moptipy.api.space.Space` of bin game plans.

The bin game plans object is defined in module
:mod:`~moptipyapps.ttp.game_plan`. Basically, it is a
two-dimensional numpy array holding, for each day (or time slot) for each team
the opposing team.
"""
from typing import Final

import numpy as np
from moptipy.api.space import Space
from moptipy.utils.logger import CSV_SEPARATOR, KeyValueLogSection
from pycommons.types import type_error

from moptipyapps.shared import SCOPE_INSTANCE
from moptipyapps.ttp.game_plan import GamePlan
from moptipyapps.ttp.instance import Instance


class GamePlanSpace(Space):
    """An implementation of the `Space` API of for game plans."""

    def __init__(self, instance: Instance) -> None:
        """
        Create a 2D packing space.

        :param instance: the 2d bin packing instance

        >>> inst = Instance.from_resource("circ4")
        >>> space = GamePlanSpace(inst)
        >>> space.instance is inst
        True
        """
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        #: The instance to which the packings apply.
        self.instance: Final[Instance] = instance
        self.copy = np.copyto  # type: ignore
        self.to_str = GamePlan.__str__  # type: ignore

    def create(self) -> GamePlan:
        """
        Create a game plan without assigning items to locations.

        :return: the (empty, uninitialized) packing object

        >>> inst = Instance.from_resource("circ8")
        >>> space = GamePlanSpace(inst)
        >>> x = space.create()
        >>> print(inst.rounds)
        2
        >>> print(inst.n_cities)
        8
        >>> x.shape
        (14, 8)
        >>> x.instance is inst
        True
        >>> type(x)
        <class 'moptipyapps.ttp.game_plan.GamePlan'>
        """
        return GamePlan(self.instance)

    def is_equal(self, x1: GamePlan, x2: GamePlan) -> bool:
        """
        Check if two bin game plans have the same contents.

        :param x1: the first game plan
        :param x2: the second game plan
        :return: `True` if both game plans are for the same instance and have
            the same structure

        >>> inst = Instance.from_resource("circ4")
        >>> space = GamePlanSpace(inst)
        >>> y1 = space.create()
        >>> y1.fill(0)
        >>> y2 = space.create()
        >>> y2.fill(0)
        >>> space.is_equal(y1, y2)
        True
        >>> y1 is y2
        False
        >>> y1[0, 0] = 1
        >>> space.is_equal(y1, y2)
        False
        """
        return (x1.instance is x2.instance) and np.array_equal(x1, x2)

    def from_str(self, text: str) -> GamePlan:
        """
        Convert a string to a packing.

        :param text: the string
        :return: the packing

        >>> inst = Instance.from_resource("circ6")
        >>> space = GamePlanSpace(inst)
        >>> y1 = space.create()
        >>> y1.fill(0)
        >>> y2 = space.from_str(space.to_str(y1))
        >>> space.is_equal(y1, y2)
        True
        >>> y1 is y2
        False
        """
        if not isinstance(text, str):
            raise type_error(text, "packing text", str)

        # we only want the very first line
        text = text.lstrip()
        lb: int = text.find("\n")
        if lb > 0:
            text = text[:lb].rstrip()

        x: Final[GamePlan] = self.create()
        np.copyto(x, np.fromstring(text, dtype=x.dtype, sep=CSV_SEPARATOR)
                  .reshape(x.shape))
        self.validate(x)
        return x

    def validate(self, x: GamePlan) -> None:
        """
        Check if a game plan is an instance of the right object.

        This method performs a superficial feasibility check, as in the TTP,
        we try to find feasible game plans and may have infeasible ones. All
        we check here is that the object is of the right type and dimensions
        and that it does not contain some out-of-bounds value.

        :param x: the game plan
        :raises TypeError: if any component of the game plan is of the wrong
            type
        :raises ValueError: if the game plan is not feasible
        """
        if not isinstance(x, GamePlan):
            raise type_error(x, "x", GamePlan)
        inst: Final[Instance] = self.instance
        if inst is not x.instance:
            raise ValueError(
                f"x.instance must be {inst} but is {x.instance}.")
        if inst.game_plan_dtype is not x.dtype:
            raise ValueError(f"inst.game_plan_dtype = {inst.game_plan_dtype}"
                             f" but x.dtype={x.dtype}")

        n: Final[int] = inst.n_cities  # the number of teams
        # each team plays every other team 'rounds'  times
        n_days: Final[int] = (n - 1) * inst.rounds

        needed_shape: Final[tuple[int, int]] = (n_days, n)
        if x.shape != needed_shape:
            raise ValueError(f"x.shape={x.shape}, but must be {needed_shape}.")
        min_id: Final[int] = -n

        for i in range(n_days):
            for j in range(n):
                v = x[i, j]
                if not (min_id <= v <= n):
                    raise ValueError(f"value {v} at x[{i}, {j}] should be in "
                                     f"{min_id}...{n}, but is not.")

    def n_points(self) -> int:
        """
        Get the number of game plans.

        The values in a game plan go from `-n..n`, including zero, and we have
        `days*n` values. This gives `(2n + 1) ** (days * n)`, where `days`
        equals `(n - 1) * rounds` and `rounds` is the number of rounds. In
        total, this gives `(2n + 1) ** ((n - 1) * rounds * n)`.

        :return: the number of possible game plans

        >>> space = GamePlanSpace(Instance.from_resource("circ6"))
        >>> print((2 * 6 + 1) ** ((6 - 1) * 2 * 6))
        6864377172744689378196133203444067624537070830997366604446306636401
        >>> space.n_points()
        6864377172744689378196133203444067624537070830997366604446306636401
        >>> space = GamePlanSpace(Instance.from_resource("circ4"))
        >>> space.n_points()
        79766443076872509863361
        >>> print((2 * 4 + 1) ** ((4 - 1) * 2 * 4))
        79766443076872509863361
        """
        inst: Final[Instance] = self.instance
        n: Final[int] = inst.n_cities
        n_days: Final[int] = (n - 1) * inst.rounds
        total_values: Final[int] = 2 * n + 1
        return total_values ** (n_days * n)

    def __str__(self) -> str:
        """
        Get the name of the game plan space.

        :return: the name, simply `gp_` + the instance name

        >>> print(GamePlanSpace(Instance.from_resource("bra24")))
        gp_bra24
        """
        return f"gp_{self.instance}"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the space to the given logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope(SCOPE_INSTANCE) as kv:
            self.instance.log_parameters_to(kv)



"""
A linear controller.

In a linear controller, all values of the state vector enter only as-is,
i.e., it is a linear combination of A and B if the state is composed of
the two values A and B. We then optimize the weights of these coefficients.
"""

from typing import Final

import numba  # type: ignore
import numpy as np

from moptipyapps.dynamic_control.controller import Controller
from moptipyapps.dynamic_control.system import System


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __linear_2d_1o(state: np.ndarray, _: float,
                   params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a linear function for 2d state spaces.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    out[0] = (state[0] * params[0]) + (state[1] * params[1])


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __linear_3d_1o(state: np.ndarray, _: float,
                   params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a linear function for 3d state spaces.

    :param state: the current state of the system
    :param params: the parameters of the polynomial.
    :param out: the control vector, receiving one single element
    """
    out[0] = (state[0] * params[0]) + (state[1] * params[1]) \
        + (state[2] * params[2])


def linear(system: System) -> Controller:
    """
    Create a linear controller for the given equations object.

    :param system: the equations object
    :return: the linear controller
    """
    if system.control_dims != 1:
        raise ValueError("invalid controller dimensions "
                         f"{system.control_dims} for {system!r}.")
    name: Final[str] = "linear"
    if system.state_dims == 2:
        return Controller(name, 2, 1, 2, __linear_2d_1o)
    if system.state_dims == 3:
        return Controller(name, 3, 1, 3, __linear_3d_1o)
    raise ValueError("invalid state dimensions "
                     f"{system.state_dims} for {system!r}.")



"""
Peak functions.

Instead of synthesizing :mod:`~moptipyapps.dynamic_control.controllers.ann`
with `arctan` as activation function, we here use `exp(-a²)` as activation
function. This function does not represent a transition from -1 to 1, but a
single peak around 0.
"""

from typing import Final, Iterable

import numba  # type: ignore
import numpy as np

from moptipyapps.dynamic_control.controller import Controller
from moptipyapps.dynamic_control.system import System


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __peak(a: float) -> float:
    """
    Compute the peak function.

    :param a: the input
    :return: the output
    """
    return np.exp(-(a * a))


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __peaks_2d_1o_1(state: np.ndarray, _: float,
                    params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a single peak for 2d-spaces.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    out[0] = params[0] * __peak(params[1] + (params[2] * state[0])
                                + (params[3] * state[1]))


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __peaks_3d_1o_1(state: np.ndarray, _: float,
                    params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a single peak for 3d-spaces.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    out[0] = params[0] * __peak(
        params[1] + (params[2] * state[0]) + (params[3] * state[1])
        + (params[4] * state[2]))


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __peaks_2d_1o_2(state: np.ndarray, _: float,
                    params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute two peaks for 2d-spaces.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    out[0] = params[0] * __peak(params[1] + (params[2] * s0)
                                + (params[3] * s1)) \
        + params[4] * __peak(params[5] + (params[6] * s0)
                             + (params[7] * s1))


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __peaks_3d_1o_2(state: np.ndarray, _: float,
                    params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a two peaks for 3d-spaces.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    s2: Final[float] = state[2]
    out[0] = params[0] * __peak(
        params[1] + (params[2] * s0) + (params[3] * s1) + (params[4] * s2))\
        + params[5] * __peak(
        params[6] + (params[7] * s0) + (params[8] * s1) + (params[9] * s2))


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __peaks_2d_1o_3(state: np.ndarray, _: float,
                    params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute three peaks for 2d-spaces.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    out[0] = params[0] * __peak(params[1] + (params[2] * s0)
                                + (params[3] * s1)) \
        + params[4] * __peak(params[5] + (params[6] * s0)
                             + (params[7] * s1)) \
        + params[8] * __peak(params[9] + (params[10] * s0)
                             + (params[11] * s1))


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def __peaks_3d_1o_3(state: np.ndarray, _: float,
                    params: np.ndarray, out: np.ndarray) -> None:
    """
    Compute a single peak for 3d-spaces.

    :param state: the current state of the system
    :param params: the weight vector for the neurons.
    :param out: the control vector, receiving one single element
    """
    s0: Final[float] = state[0]
    s1: Final[float] = state[1]
    s2: Final[float] = state[2]
    out[0] = params[0] * __peak(
        params[1] + (params[2] * s0) + (params[3] * s1) + (params[4] * s2))\
        + params[5] * __peak(
        params[6] + (params[7] * s0) + (params[8] * s1) + (params[9] * s2))\
        + params[10] * __peak(
        params[11] + (params[12] * s0) + (params[13] * s1)
        + (params[14] * s2))


def peaks(system: System) -> Iterable[Controller]:
    """
    Create poor man's PNNs fitting to a given system.

    Based on the dimensionality of the state space, we generate a set of PNNs
    with different numbers of layers and neurons. The weights of the neurons
    can then directly be optimized by a numerical optimization algorithm.
    This is, of course, probably much less efficient that doing some proper
    learning like back-propagation. However, it allows us to easily plug the
    PNNs into the same optimization routines as other controllers.

    :param system: the equations object
    :return: the PNNs
    """
    if system.control_dims != 1:
        raise ValueError("invalid controller dimensions "
                         f"{system.control_dims} for {system!r}.")
    if system.state_dims == 2:
        return (Controller("peaks_1", 2, 1, 4, __peaks_2d_1o_1),
                Controller("peaks_2", 2, 1, 8, __peaks_2d_1o_2),
                Controller("peaks_3", 2, 1, 12, __peaks_2d_1o_3))
    if system.state_dims == 3:
        return (Controller("peaks_1", 3, 1, 5, __peaks_3d_1o_1),
                Controller("peaks_2", 3, 1, 10, __peaks_3d_1o_2),
                Controller("peaks_3", 3, 1, 15, __peaks_3d_1o_3))
    raise ValueError("invalid state dimensions "
                     f"{system.state_dims} for {system!r}.")



"""
An objective function counting deviations from the instance definition.

This objective function will be the smaller, the closer the structure
of an instance is to the original instance.
Due to our encoding, we create instances whose bin width, bin height,
and the number of items is the same as in an existing instance. The
lower bound for the required number of bins is also the same.

This objective function here also incorporates some additional features, such
as:

1. Is the number of different items similar to those in the original instance?
   In an existing instance, some items of same width and height could be
   grouped together. We may have 10 items to pack, but only 3 different item
   sizes exist. We here compare the number of different item sizes of a
   generated instance with the number in the instance definition.
2. In a given instance, we can observe the minimum and maximum width and
   height of any item. If an item in the generated instance is smaller than
   the minimum or larger than the maximum permitted value in one dimension,
   this will increase the error count.
3. Additionally, we want the actual minimum and maximum width and height of
   any item in the generated instance is the same as in the original instance.
4. Finally, we want that the total area covered by all items is the same in
   the generated instance as in the original instance.

All of these potential violations are counted and added up. Using this
objective function should drive the search towards generating instances that
are structurally similar to an existing instance, at least in some parameters.

We could push this arbitrarily further, trying to emulate the exact
distribution of the item sizes, etc. But this may just lead to the
reproduction of the original instance by the search and not add anything
interesting.

>>> orig = Instance.from_resource("a04")
>>> space = InstanceSpace(orig)
>>> print(f"{space.inst_name!r} with {space.n_different_items}/"
...       f"{space.n_items} items with area {space.total_item_area} "
...       f"in {space.min_bins} bins of "
...       f"size {space.bin_width}*{space.bin_height}.")
'a04n' with 2/16 items with area 7305688 in 3 bins of size 2750*1220.

>>> from moptipyapps.binpacking2d.instgen.inst_decoding import InstanceDecoder
>>> decoder = InstanceDecoder(space)
>>> import numpy as np
>>> x = np.array([ 0.0,  0.2, -0.1,  0.3,  0.5, -0.6, -0.7,  0.9,
...                0.0,  0.2, -0.1,  0.3,  0.5, -0.6, -0.7,  0.9,
...                0.0,  0.2, -0.1,  0.3,  0.5, -0.6, -0.7,  0.9,
...                0.0,  0.2, ])
>>> y = space.create()
>>> decoder.decode(x, y)
>>> space.validate(y)
>>> res: Instance = y[0]
>>> print(f"{res.name!r} with {res.n_different_items}/"
...       f"{res.n_items} items with area {res.total_item_area} "
...       f"in {res.lower_bound_bins} bins of "
...       f"size {res.bin_width}*{res.bin_height}.")
'a04n' with 15/16 items with area 10065000 in 3 bins of size 2750*1220.
>>> print(space.to_str(y))
a04n;15;2750;1220;1101,1098;2750,244;2750,98;1101,171;1649,171;2750,976;\
441,122;1649,122;2750,10;2750,1,2;2750,3;1649,1098;2750,878;2750,58;660,122

>>> errors = Errors(space)
>>> errors.lower_bound()
0
>>> errors.upper_bound()
7332043
>>> errors.evaluate(y)
2770589

>>> y[0] = orig
>>> errors.evaluate(y)
0
"""
from typing import Final

from moptipy.api.objective import Objective
from pycommons.types import type_error

from moptipyapps.binpacking2d.instance import (
    IDX_HEIGHT,
    IDX_REPETITION,
    IDX_WIDTH,
    Instance,
)
from moptipyapps.binpacking2d.instgen.instance_space import InstanceSpace


class Errors(Objective):
    """Compute the deviation of an instance from the definition."""

    def __init__(self, space: InstanceSpace) -> None:
        """
        Initialize the errors objective function.

        :param instance: the instance to load the bounds from
        """
        super().__init__()
        if not isinstance(space, InstanceSpace):
            raise type_error(space, "space", InstanceSpace)
        #: the instance description
        self.space: Final[InstanceSpace] = space

    def evaluate(self, x: list[Instance] | Instance) -> int:
        """
        Compute the deviations from the instance definition.

        :param x: the instance
        :return: the number of deviations
        """
        errors: int = 0
        inst: Final[Instance] = x[0] if isinstance(x, list) else x
        space: Final[InstanceSpace] = self.space

        # Some errors are not permitted.
        errors += abs(inst.bin_width - space.bin_width)  # should be 0!
        errors += abs(inst.bin_height - space.bin_height)  # should be 0!
        errors += abs(inst.n_items - space.n_items)  # should be 0!
        if errors > 0:
            raise ValueError(
                f"Instance {inst} is inconsistent with space {space}.")

        # These errors are permitted.
        n_different: Final[int] = inst.n_different_items
        errors += abs(n_different - space.n_different_items)  # > 0
        goal_min_width: Final[int] = space.item_width_min
        goal_max_width: Final[int] = space.item_width_max
        goal_min_height: Final[int] = space.item_height_min
        goal_max_height: Final[int] = space.item_height_max

        actual_min_width: int = space.bin_width
        actual_max_width: int = 0
        actual_min_height: int = space.bin_height
        actual_max_height: int = 0
        total_area: int = 0

        for row in range(n_different):
            width: int = int(inst[row, IDX_WIDTH])
            height: int = int(inst[row, IDX_HEIGHT])
            actual_min_width = min(actual_min_width, width)
            actual_max_width = max(actual_max_width, width)
            actual_min_height = min(actual_min_height, height)
            actual_max_height = max(actual_max_height, height)

            n: int = int(inst[row, IDX_REPETITION])
            total_area += n * width * height
            if width < goal_min_width:
                errors += n * (goal_min_width - width)
            elif width > goal_max_width:
                errors += n * (width - goal_max_width)
            if height < goal_min_height:
                errors += n * (goal_min_height - height)
            elif height > goal_max_height:
                errors += n * (height - goal_max_height)

        errors += abs(actual_min_width - goal_min_width)
        errors += abs(actual_max_width - goal_max_width)
        errors += abs(actual_min_height - goal_min_height)
        errors += abs(actual_max_height - goal_max_height)

        if total_area != inst.total_item_area:
            raise ValueError("Invalid total area?")
        errors += abs(total_area - space.total_item_area)

        return errors

    def lower_bound(self) -> int:
        """
        Get the lower bound of the instance template deviations.

        :return: the lower bound for the errors
        """
        return 0

    def is_always_integer(self) -> bool:
        """
        Return `True` because there are only integer errors.

        :retval True: always
        """
        return True

    def upper_bound(self) -> int:
        """
        Get the upper bound of the number of deviations.

        :return: the upper bound
        """
        space: Final[InstanceSpace] = self.space

        # These errors are permitted.
        max_errors: int = max(
            space.n_different_items - 1,
            space.n_items - space.n_different_items - 1)

        goal_min_width: Final[int] = space.item_width_min
        goal_max_width: Final[int] = space.item_width_max
        goal_min_height: Final[int] = space.item_height_min
        goal_max_height: Final[int] = space.item_height_max
        bin_width: Final[int] = space.bin_width
        bin_height: Final[int] = space.bin_height

        n_items: Final[int] = space.n_items
        max_errors += n_items * max(
            goal_min_width - 1, bin_width - goal_max_width)
        max_errors += n_items * max(
            goal_min_height - 1, bin_height - goal_max_height)
        max_errors += max(goal_min_width, bin_width - goal_min_width)
        max_errors += max(goal_max_width, bin_width - goal_max_width)
        max_errors += max(goal_min_height, bin_height - goal_min_height)
        max_errors += max(goal_max_height, bin_height - goal_max_height)

        alt_area: Final[int] = (space.min_bins * bin_width * bin_height
                                - space.total_item_area)
        if alt_area < 0:
            raise ValueError("Invalid item area in space?")
        max_errors += max(space.total_item_area, alt_area)

        return max_errors

    def __str__(self) -> str:
        """
        Get the name of the errors objective function.

        :return: `errors`
        :retval "errors": always
        """
        return "errors"



"""
A combination of the errors and the hardness objective.

>>> orig = Instance.from_resource("a04")
>>> space = InstanceSpace(orig)
>>> print(f"{space.inst_name!r} with {space.n_different_items}/"
...       f"{space.n_items} items with area {space.total_item_area} "
...       f"in {space.min_bins} bins of "
...       f"size {space.bin_width}*{space.bin_height}.")
'a04n' with 2/16 items with area 7305688 in 3 bins of size 2750*1220.

>>> from moptipyapps.binpacking2d.instgen.inst_decoding import InstanceDecoder
>>> decoder = InstanceDecoder(space)
>>> import numpy as np
>>> x = np.array([ 0.0,  0.2, -0.1,  0.3,  0.5, -0.6, -0.7,  0.9,
...                0.0,  0.2, -0.1,  0.3,  0.5, -0.6, -0.7,  0.9,
...                0.0,  0.2, -0.1,  0.3,  0.5, -0.6, -0.7,  0.9,
...                0.0,  0.2, ])
>>> y = space.create()
>>> decoder.decode(x, y)
>>> space.validate(y)
>>> res: Instance = y[0]
>>> print(f"{res.name!r} with {res.n_different_items}/"
...       f"{res.n_items} items with area {res.total_item_area} "
...       f"in {res.lower_bound_bins} bins of "
...       f"size {res.bin_width}*{res.bin_height}.")
'a04n' with 15/16 items with area 10065000 in 3 bins of size 2750*1220.
>>> print(space.to_str(y))
a04n;15;2750;1220;1101,1098;2750,244;2750,98;1101,171;1649,171;2750,976;\
441,122;1649,122;2750,10;2750,1,2;2750,3;1649,1098;2750,878;2750,58;660,122

>>> obj = ErrorsAndHardness(space, max_fes=100)
>>> obj.lower_bound()
1.0
>>> obj.upper_bound()
14664088.0
>>> obj.evaluate(y)
4132864.157998087

>>> obj.evaluate(orig)
1.9606359978301315
"""
from typing import Callable, Final, Iterable

from moptipy.api.execution import Execution
from moptipy.api.objective import Objective
from moptipy.utils.logger import KeyValueLogSection

from moptipyapps.binpacking2d.instance import (
    Instance,
)
from moptipyapps.binpacking2d.instgen.errors import Errors
from moptipyapps.binpacking2d.instgen.hardness import (
    DEFAULT_EXECUTORS,
    Hardness,
)
from moptipyapps.binpacking2d.instgen.instance_space import InstanceSpace


class ErrorsAndHardness(Objective):
    """Compute the errors and hardness."""

    def __init__(self, space: InstanceSpace,
                 max_fes: int = 1_000_000, n_runs: int = 3,
                 executors: Iterable[
                     Callable[[Instance], tuple[
                         Execution, Objective]]] = DEFAULT_EXECUTORS) -> None:
        """
        Initialize the errors objective function.

        :param space: the instance space
        :param max_fes: the maximum FEs
        :param n_runs: the maximum runs
        :param executors: the functions creating the executions
        """
        super().__init__()
        #: the errors objective
        self.errors: Final[Errors] = Errors(space)
        #: the hardness objective function
        self.hardness: Final[Hardness] = Hardness(
            max_fes, n_runs, executors)

    def evaluate(self, x: list[Instance] | Instance) -> float:
        """
        Compute the combination of hardness and errors.

        :param x: the instance
        :return: the hardness and errors
        """
        return (1 + self.errors.evaluate(x)) * (
            1.0 + self.hardness.evaluate(x))

    def lower_bound(self) -> float:
        """
        Get the lower bound of the instance error and hardness.

        :return: the lower bound for the hardness and errors
        """
        return (1 + self.errors.lower_bound()) * (
            1.0 + self.hardness.lower_bound())

    def upper_bound(self) -> float:
        """
        Get the upper bound of the instance error and hardness.

        :return: the upper bound for the instance hardness and errors
        """
        return (1 + self.errors.upper_bound()) * (
            1.0 + self.hardness.upper_bound())

    def is_always_integer(self) -> bool:
        """
        Return `False` because the hardness function returns `float`.

        :retval False: always
        """
        return False

    def __str__(self) -> str:
        """
        Get the name of the errors objective function.

        :return: `errorsAndHardness`
        :retval "errorsAndHardness": always
        """
        return "errorsAndHardness"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of this instance.

        :param logger: the logger
        """
        super().log_parameters_to(logger)
        with logger.scope("err") as err:
            self.errors.log_parameters_to(err)
        with logger.scope("hard") as hard:
            self.hardness.log_parameters_to(hard)



"""
An instance of the dynamic control synthesis problem.

An instance of the dynamic control synthesis problem is comprised of two
components: a :mod:`~moptipyapps.dynamic_control.system` of differential
equations governing how the state of a system changes over time and a
:mod:`~moptipyapps.dynamic_control.controller` that uses the current system
state as input and computes a controller value as output that influences the
state change. Notice that the controller here is a parametric function. The
goal of the dynamic system control is to find the right parameterization of
the controller such that an :mod:`~moptipyapps.dynamic_control.objective` is
minimized. The objective here usually has the goal to bring the dynamic system
into a stable state while using as little controller "energy" as possible.

An instance of the simultaneous control and model synthesis problem is an
instance of the class :class:`~moptipyapps.dynamic_control.system_\
model.SystemModel`, which is a subclass of :class:`~moptipyapps.\
dynamic_control.instance.Instance`. It also adds a controller blueprint for
modelling the systems response (state differential) based on the system state
and controller output.

The starting point of the work here were conversations with Prof. Dr. Bernd
NOACK and Guy Yoslan CORNEJO MACEDA of the Harbin Institute of Technology in
Shenzhen, China (哈尔滨工业大学(深圳)).
"""

from typing import Final

import numpy as np
from moptipy.api.component import Component
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.strings import sanitize_name
from pycommons.io.path import Path
from pycommons.types import type_error

from moptipyapps.dynamic_control.controller import Controller
from moptipyapps.dynamic_control.system import System


class Instance(Component):
    """An instance of the dynamic control problem."""

    def __init__(self, system: System, controller: Controller,
                 name_base: str | None = None) -> None:
        """
        Create an instance of the dynamic control problem.

        :param system: the system of equations governing the dynamic system
        :param controller: the controller applied to the system
        :param name_base: the name base
        """
        super().__init__()
        if not isinstance(system, System):
            raise type_error(system, "system", System)
        if not isinstance(controller, Controller):
            raise type_error(controller, "controller", Controller)
        if controller.state_dims != system.state_dims:
            raise ValueError(
                f"controller.state_dims={controller.state_dims}, but "
                f"system.state_dims={system.state_dims} for controller "
                f"{str(controller)!r} and system {str(system)!r}.")
        if controller.control_dims != system.control_dims:
            raise ValueError(
                f"controller.control_dims={controller.control_dims}, but "
                f"system.control_dims={system.control_dims} for controller "
                f"{str(controller)!r} and system {str(system)!r}.")
        #: the system governing the dynamic system
        self.system: Final[System] = system
        #: the controller applied to the system
        self.controller: Final[Controller] = controller

        name: str = f"{self.system}_{self.controller}"
        if name_base is not None:
            if not isinstance(name_base, str):
                raise type_error(name_base, "name_base", (str, None))
            nn = sanitize_name(name_base)
            if nn != name_base:
                raise ValueError(f"sanitized name base {nn!r} different "
                                 f"from original {name_base!r}.")
            name = f"{name}_{name_base}"
        #: the name of this instance
        self.name: Final[str] = name

    def __str__(self) -> str:
        """
        Get the name of this instance.

        :return: a combination of the equation name and the controller name
        """
        return self.name

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope("sys") as scope:
            self.system.log_parameters_to(scope)
        with logger.scope("ctrl") as scope:
            self.controller.log_parameters_to(scope)

    def describe_parameterization(
            self, title: str | None,
            parameters: np.ndarray, base_name: str,
            dest_dir: str) -> tuple[Path, ...]:
        """
        Describe the performance of a given system of system.

        :param title: the optional title
        :param parameters: the controller parameters
        :param base_name: the base name of the file to produce
        :param dest_dir: the destination directory
        :returns: the paths of the generated files
        """
        the_title = self.controller.name
        if title is not None:
            the_title = f"{the_title}\n{title}"
        return self.system.describe_system(
            the_title, self.controller.controller, parameters,
            base_name, dest_dir)
